# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import LevenbergMarquardt
from nemos.third_party.jaxopt.jaxopt._src import linear_solve
from nemos.third_party.jaxopt.jaxopt._src import test_util

import numpy as onp


def _rosenbrock(x):
  return jnp.sqrt(2) * jnp.array([10 * (x[1] - x[0]**2), (1 - x[0])])


def _scaled_meyer_residual(coeffs, x, y):
  return 1e-3 * y - coeffs[0] * jnp.exp(10 * coeffs[1] / (x + coeffs[2]) - 13)


def _double_exponential(coeffs, x):
  return coeffs[0] * jnp.exp(-coeffs[2] * x) + coeffs[1] * jnp.exp(
      -coeffs[3] * x)


def _double_exponential_residual(coeffs, x, y):
  return y - _double_exponential(coeffs, x)


def _enzyme_reaction_residual_model(coeffs, x, y):
  return y - coeffs[0] * x / (coeffs[1] + x)


def _enzyme_reaction_residual_model_jac(coeffs, x, y, eps=1e-5):
  """Return the numerical Jacobian."""
  lm = LevenbergMarquardt(
      residual_fun=_enzyme_reaction_residual_model,)

  # Sets eps only at idx, the rest is zero
  eps_at = lambda idx: onp.array([int(i == idx)*eps for i in range(len(x))])

  res1 = jnp.zeros((len(coeffs), len(x)))
  res2 = jnp.zeros((len(coeffs), len(x)))
  for i in range(len(x)):
    res1 = res1.at[:, i].set(lm.run(coeffs, x + eps_at(i), y).params)
    res2 = res2.at[:, i].set(lm.run(coeffs, x - eps_at(i), y).params)

  twoeps = 2 * eps
  return (res1 - res2) / twoeps


def _custom_solve_inv(matvec, b, ridge, init=None):
  matvec = linear_solve._make_ridge_matvec(matvec, ridge=ridge)
  A = linear_solve._materialize_array(matvec, b.shape)
  return jnp.dot(jnp.linalg.inv(A), b)

# Uncomment this line to test in x64 
# Note: this line needs to be defined before the Test Class
# jax.config.update('jax_enable_x64', True)

class LevenbergMarquardtTest(test_util.JaxoptTestCase):

  def setUp(self):
    super().setUp()

    self.substrate_conc = onp.array(
        [0.038, 0.194, .425, .626, 1.253, 2.500, 3.740])
    self.rate_data = onp.array(
        [0.050, 0.127, 0.094, 0.2122, 0.2729, 0.2665, 0.3317])
    self.init_enzyme_reaction_coeffs = onp.array([0.1, 0.1])

    # Exponential fit data from immoptibox "A MATLAB TOOLBOX FOR OPTIMIZATION
    # AND DATA FITTING" by Hans Bruun Nielsen
    self.double_exponential_x = onp.linspace(0.02, 0.9, 45)
    self.double_exponential_y = onp.array([
        0.090542, 0.124569, 0.179367, 0.195654, 0.269707, 0.286027, 0.289892,
        0.317475, 0.308191, 0.336995, 0.348371, 0.321337, 0.299423, 0.338972,
        0.304763, 0.288903, 0.30082, 0.303974, 0.283987, 0.262078, 0.281593,
        0.267531, 0.218926, 0.225572, 0.200594, 0.197375, 0.18244, 0.183892,
        0.152285, 0.174028, 0.150874, 0.12622, 0.126266, 0.106384, 0.118923,
        0.091868, 0.128926, 0.119273, 0.115997, 0.105831, 0.075261, 0.068387,
        0.090823, 0.085205, 0.067203
    ])

    # Meyers fit data from example 6.21 of K. Madsen & H. B. Nielsen in the 
    # book "Introduction to Optimization and Data Fitting".
    self.scaled_meyer_x = onp.linspace(0.5, 1.25, 16)
    self.scaled_meyer_y = onp.array([
        34780, 28610, 23650, 19630, 16370, 13720, 11540, 9744, 8261, 7030,
        6005, 5147, 4427, 3820, 3307, 2872
    ])

  @unittest.skipIf(jax.config.jax_enable_x64, 'test requires X32')
  def test_rosenbrock_x32(self):
    lm = LevenbergMarquardt(
        residual_fun=_rosenbrock,
        damping_parameter=1e-3,
        maxiter=16)
    optimize_info = lm.run(onp.array([-1.2, 1]))

    self.assertArraysAllClose(optimize_info.params,
                              onp.array([0.999948, 0.999894], float))
    self.assertAllClose(optimize_info.state.iter_num, 14)

  @unittest.skipIf(jax.config.jax_enable_x64, 'test requires X32')
  @parameterized.product(materialize_jac=[True, False])
  def test_rosenbrock_x32_materialize_jac(self, materialize_jac):

    lm = LevenbergMarquardt(
        residual_fun=_rosenbrock,
        damping_parameter=1e-3,
        maxiter=16,
        materialize_jac=materialize_jac)
    optimize_info = lm.run(onp.array([-1.2, 1]))

    self.assertArraysAllClose(optimize_info.params,
                              onp.array([0.999948, 0.999894], float))
    self.assertAllClose(optimize_info.state.iter_num, 14)

  # This unit test is for checking rosenbrock function based on the example
  # provided by K. Madsen & H. B. Nielsen in the book "Introduction to
  # Optimization and Data Fitting" Example 6.9. page 123.
  @unittest.skipIf(not jax.config.jax_enable_x64, 'test requires X64')
  def test_rosenbrock_x64(self):
    lm = LevenbergMarquardt(
        residual_fun=_rosenbrock,
        damping_parameter=1e-3,
        stop_criterion='madsen-nielsen',
        xtol=1e-12,
        gtol=1e-8,
        maxiter=16)
    optimize_info = lm.run(onp.array([-1.2, 1]))

    self.assertArraysAllClose(
        optimize_info.params - onp.array([1, 1]),
        onp.array([-4.07e-09, -8.16e-09]),
        atol=1e-11)
    self.assertAllClose(optimize_info.state.iter_num, 16)

  @unittest.skipIf(not jax.config.jax_enable_x64, 'test requires X64')
  @parameterized.product(materialize_jac=[True, False])
  def test_rosenbrock_x64_materialize_jac(self, materialize_jac):
    lm = LevenbergMarquardt(
        residual_fun=_rosenbrock,
        damping_parameter=1e-3,
        stop_criterion='madsen-nielsen',
        xtol=1e-12,
        gtol=1e-8,
        maxiter=16,
        materialize_jac=materialize_jac)
    optimize_info = lm.run(onp.array([-1.2, 1]))

    self.assertArraysAllClose(
        optimize_info.params - onp.array([1, 1]),
        onp.array([-4.07e-09, -8.16e-09]),
        atol=1e-11)
    self.assertAllClose(optimize_info.state.iter_num, 16)

  @unittest.skipIf(jax.config.jax_enable_x64, 'test requires X32')
  @parameterized.product(materialize_jac=[True, False], 
                         solver=['cholesky', 'lu', 'qr', 'inv'])
  def test_scaled_meyer_x32(self, materialize_jac, solver):
    if materialize_jac == False and solver == 'qr':
      self.skipTest("QR factorization solver requires materialization of Jacobian.")
    
    lm = LevenbergMarquardt(
        residual_fun=_scaled_meyer_residual,
        damping_parameter=10,
        tol=1e-3,
        maxiter=200,
        solver=solver,
        materialize_jac=materialize_jac,
        geodesic=True)
    optimize_info = lm.run(
        onp.asarray([40, 50, 30], dtype=float),
        self.scaled_meyer_x,
        self.scaled_meyer_y)

    self.assertArraysAllClose(
        optimize_info.params, onp.array([2.48, 6.18, 3.45], dtype=float), atol=5e-2, rtol=2e-2)
    
  @unittest.skipIf(not jax.config.jax_enable_x64, 'test requires X64')
  @parameterized.product(materialize_jac=[True, False], 
                         solver=['cholesky', 'lu', 'qr', 'inv'])
  def test_scaled_meyer_x64(self, materialize_jac, solver):
    if materialize_jac == False and solver == 'qr':
      self.skipTest("QR factorization solver requires materialization of Jacobian.")
    
    lm = LevenbergMarquardt(
        residual_fun=_scaled_meyer_residual,
        damping_parameter=10,
        tol=1e-3,
        maxiter=200,
        solver=solver,
        materialize_jac=materialize_jac,
        geodesic=True)
    optimize_info = lm.run(
        onp.asarray([40, 50, 30], dtype=float),
        self.scaled_meyer_x,
        self.scaled_meyer_y)

    self.assertArraysAllClose(
        optimize_info.params, onp.array([2.48, 6.18, 3.45], dtype=float), rtol=5*1e-3)

  # This unit test is for checking rosenbrock function based on the example
  # provided by K. Madsen & H. B. Nielsen in the book "Introduction to
  # Optimization and Data Fitting" Example 6.10-. page 124. Data taken from
  # immoptibox "A MATLAB TOOLBOX FOR OPTIMIZATION AND DATA FITTING" H.B. Nielsen
  @unittest.skipIf(not jax.config.jax_enable_x64, 'test requires X64')
  @parameterized.product(materialize_jac=[True, False],
                         solver=['cholesky', 'lu', 'qr', 'inv', _custom_solve_inv],
                         geodesic=[True, False],)
  def test_double_exponential_fit_x64(self, materialize_jac, solver, geodesic):
    if materialize_jac == False and solver == 'qr':
      self.skipTest("QR factorization solver requires materialization of Jacobian.")
    lm = LevenbergMarquardt(
        residual_fun=_double_exponential_residual,
        damping_parameter=1e-3,
        stop_criterion='madsen-nielsen',
        xtol=1e-14,
        gtol=1e-8,
        maxiter=100,
        solver=solver,
        geodesic=geodesic,
        materialize_jac=materialize_jac)
    optimize_info = lm.run(
        onp.asarray([1, -1, 1, 2], dtype=float),
        self.double_exponential_x,
        self.double_exponential_y)

    # Check the final result up to 4 sig digits.
    self.assertArraysAllClose(
        optimize_info.params, onp.array([4, -4, 4, 5], dtype=float), atol=5*1e-3)

  # TODO(vroulet) This test is failing on github, producing NaNs. Needs to fix algo.
  # @parameterized.product(materialize_jac=[True, False],
  #                        solver=['cholesky', 'lu', 'qr', 'inv', _custom_solve_inv],
  #                        geodesic=[True, False],)
  # def test_double_exponential_fit_x32(self, materialize_jac, solver, geodesic):
  #   if materialize_jac == False and solver == 'qr':
  #     self.skipTest("QR factorization solver requires materialization of Jacobian.")
  #   lm = LevenbergMarquardt(
  #       residual_fun=_double_exponential_residual,
  #       damping_parameter=1e-3,
  #       stop_criterion='madsen-nielsen',
  #       xtol=1e-6,
  #       gtol=1e-6,
  #       maxiter=200,
  #       solver=solver,
  #       geodesic=geodesic,
  #       materialize_jac=materialize_jac)
  #   optimize_info = lm.run(
  #       onp.asarray([1, -1, 1, 2], dtype=float),
  #       self.double_exponential_x,
  #       self.double_exponential_y)

  #   # Check the final result up to 4 sig digits.
  #   self.assertArraysAllClose(
  #       optimize_info.params, onp.array([4, -4, 4, 5], dtype=float), atol=0.5)

  @parameterized.product(implicit_diff=[True, False])
  def test_enzyme_reaction_implicit_diff(self, implicit_diff):
    jac_num = _enzyme_reaction_residual_model_jac(
        self.init_enzyme_reaction_coeffs, self.substrate_conc, self.rate_data)

    lm = LevenbergMarquardt(
        residual_fun=_enzyme_reaction_residual_model,
        tol=1e-6,
        implicit_diff=implicit_diff)

    def wrapper(substrate_conc):
      return lm.run(
          self.init_enzyme_reaction_coeffs,
          substrate_conc,
          self.rate_data).params
    jac_custom = jax.jacrev(wrapper)(self.substrate_conc)

    self.assertArraysAllClose(jac_num, jac_custom, atol=5e-2)

  def test_has_aux_false(self):
    lm = LevenbergMarquardt(
        residual_fun=lambda x, y: jnp.abs(x - y), tol=1e-6, has_aux=False)
    x_init = jnp.zeros((10,))
    x_gt = jnp.linspace(0, 5, 10)
    x_opt, state = lm.run(x_init, x_gt)

    self.assertAllClose(x_opt, x_gt, atol=1e-6)
    self.assertIsNone(state.aux)

  def test_has_aux_true(self):
    lm = LevenbergMarquardt(
        residual_fun=lambda x, y: (jnp.abs(x - y), x**2),
        tol=1e-6,
        has_aux=True)
    x_init = jnp.zeros((10,))
    x_gt = jnp.linspace(0, 5, 10)
    x_opt, state = lm.run(x_init, x_gt)

    self.assertAllClose(x_opt, x_gt, atol=1e-6)
    self.assertAllClose(state.aux, x_gt**2, atol=1e-6)

  def test_scalar_output_fun(self):
    lm = LevenbergMarquardt(
        residual_fun=lambda x: x @ x,
        tol=1e-1,)
    x_init = jnp.ones((2,))
    x_opt, _ = lm.run(x_init)

    self.assertAllClose(x_opt, jnp.zeros((2,)), atol=1e0)

if __name__ == '__main__':
  absltest.main()



import itertools
from contextlib import nullcontext as does_not_raise

import jax.numpy as jnp
import numpy as np
import pytest

import nemos.simulation as simulation
from nemos import basis


@pytest.mark.parametrize(
    "inhib_a, expectation",
    [
        (
            -1,
            pytest.raises(ValueError, match="Gamma parameter [a-z]+_[a,b] must be >0."),
        ),
        (
            0,
            pytest.raises(ValueError, match="Gamma parameter [a-z]+_[a,b] must be >0."),
        ),
        (1, does_not_raise()),
    ],
)
def test_difference_of_gammas_inhib_a(inhib_a, expectation):
    with expectation:
        simulation.difference_of_gammas(10, inhib_a=inhib_a)


@pytest.mark.parametrize(
    "excit_a, expectation",
    [
        (
            -1,
            pytest.raises(ValueError, match="Gamma parameter [a-z]+_[a,b] must be >0."),
        ),
        (
            0,
            pytest.raises(ValueError, match="Gamma parameter [a-z]+_[a,b] must be >0."),
        ),
        (1, does_not_raise()),
    ],
)
def test_difference_of_gammas_excit_a(excit_a, expectation):
    with expectation:
        simulation.difference_of_gammas(10, excit_a=excit_a)


@pytest.mark.parametrize(
    "upper_percentile, expectation",
    [
        (
            -0.1,
            pytest.raises(
                ValueError,
                match=r"upper_percentile should lie in the \[0, 1\) interval.",
            ),
        ),
        (0, does_not_raise()),
        (0.1, does_not_raise()),
        (
            1,
            pytest.raises(
                ValueError,
                match=r"upper_percentile should lie in the \[0, 1\) interval.",
            ),
        ),
        (
            10,
            pytest.raises(
                ValueError,
                match=r"upper_percentile should lie in the \[0, 1\) interval.",
            ),
        ),
    ],
)
def test_difference_of_gammas_percentile_params(upper_percentile, expectation):
    with expectation:
        simulation.difference_of_gammas(10, upper_percentile)


@pytest.mark.parametrize("window_size", [0, 1, 2])
def test_difference_of_gammas_output_shape(window_size):
    result_size = simulation.difference_of_gammas(window_size).size
    assert (
        result_size == window_size
    ), f"Expected output size {window_size}, but got {result_size}"


@pytest.mark.parametrize("window_size", [1, 2, 10])
def test_difference_of_gammas_output_norm(window_size):
    result = simulation.difference_of_gammas(window_size)
    assert np.allclose(
        np.linalg.norm(result, ord=2), 1
    ), "The output of difference_of_gammas is not unit norm."


@pytest.mark.parametrize(
    "coupling_filters, expectation",
    [
        (
            np.zeros((10,)),
            pytest.raises(
                ValueError, match=r"coupling_filters must be a 3 dimensional array"
            ),
        ),
        (
            np.zeros((10, 2)),
            pytest.raises(
                ValueError, match=r"coupling_filters must be a 3 dimensional array"
            ),
        ),
        (np.zeros((10, 2, 2)), does_not_raise()),
        (
            np.zeros((10, 2, 2, 2)),
            pytest.raises(
                ValueError, match=r"coupling_filters must be a 3 dimensional array"
            ),
        ),
    ],
)
def test_regress_filter_coupling_filters_dim(coupling_filters, expectation):
    ws = coupling_filters.shape[0]
    with expectation:
        simulation.regress_filter(coupling_filters, np.zeros((ws, 3)))


@pytest.mark.parametrize(
    "eval_basis, expectation",
    [
        (
            np.zeros((10,)),
            pytest.raises(
                ValueError, match=r"eval_basis must be a 2 dimensional array"
            ),
        ),
        (np.zeros((10, 2)), does_not_raise()),
        (
            np.zeros((10, 2, 2)),
            pytest.raises(
                ValueError, match=r"eval_basis must be a 2 dimensional array"
            ),
        ),
        (
            np.zeros((10, 2, 2, 2)),
            pytest.raises(
                ValueError, match=r"eval_basis must be a 2 dimensional array"
            ),
        ),
    ],
)
def test_regress_filter_eval_basis_dim(eval_basis, expectation):
    ws = eval_basis.shape[0]
    with expectation:
        simulation.regress_filter(np.zeros((ws, 1, 1)), eval_basis)


@pytest.mark.parametrize(
    "delta_ws, expectation",
    [
        (
            -1,
            pytest.raises(
                ValueError, match=r"window_size mismatch\. The window size of "
            ),
        ),
        (0, does_not_raise()),
        (
            1,
            pytest.raises(
                ValueError, match=r"window_size mismatch\. The window size of "
            ),
        ),
    ],
)
def test_regress_filter_window_size_matching(delta_ws, expectation):
    ws = 2
    with expectation:
        simulation.regress_filter(np.zeros((ws, 1, 1)), np.zeros((ws + delta_ws, 1)))


@pytest.mark.parametrize(
    "window_size, n_neurons_sender, n_neurons_receiver, n_basis_funcs",
    [x for x in itertools.product([1, 2], [1, 2], [1, 2], [1, 2])],
)
def test_regress_filter_weights_size(
    window_size, n_neurons_sender, n_neurons_receiver, n_basis_funcs
):
    weights = simulation.regress_filter(
        np.zeros((window_size, n_neurons_sender, n_neurons_receiver)),
        np.zeros((window_size, n_basis_funcs)),
    )
    assert weights.shape[0] == n_neurons_sender, (
        "First dimension of weights (n_neurons_receiver) does not "
        "match the second dimension of coupling_filters."
    )
    assert weights.shape[1] == n_neurons_receiver, (
        "Second dimension of weights (n_neuron_sender) does not "
        "match the third dimension of coupling_filters."
    )
    assert weights.shape[2] == n_basis_funcs, (
        "Third dimension of weights (n_basis_funcs) does not "
        "match the second dimension of eval_basis."
    )


def test_least_square_correctness():
    """
    Test the correctness of the least square estimate by enforcing an invertible map,
    i.e. a map for which the least-square estimator matches the original weights.
    """
    # set up problem dimensionality
    ws, n_neurons_receiver, n_neurons_sender, n_basis_funcs = 100, 1, 2, 10
    # evaluate a basis
    _, eval_basis = basis.RaisedCosineLinearEval(n_basis_funcs).evaluate_on_grid(ws)
    # generate random weights to define filters
    weights = np.random.normal(
        size=(n_neurons_receiver, n_neurons_sender, n_basis_funcs)
    )
    # define filters as linear combination of basis elements
    coupling_filt = np.einsum("ijk, tk -> tij", weights, eval_basis)
    # recover weights by means of linear regression
    weights_lsq = simulation.regress_filter(coupling_filt, eval_basis)
    # check the exact matching of the filters up to numerical error
    assert np.allclose(weights_lsq, weights)


class TestSimulateRecurrent:
    @pytest.mark.parametrize(
        "delta_n_neuron, expectation",
        [
            (
                -1,
                pytest.raises(ValueError, match="The number of neurons"),
            ),
            (0, does_not_raise()),
            (
                1,
                pytest.raises(ValueError, match="The number of neurons"),
            ),
        ],
    )
    def test_simulate_n_neuron_match_input(
        self, delta_n_neuron, expectation, coupled_model_simulate
    ):
        """
        Test the `simulate` method to ensure that The number of neurons in the input
        matches the model's parameters.
        """
        (
            coupling_coeff,
            feedforward_coeff,
            intercepts,
            random_key,
            feedforward_input,
            coupling_basis,
            init_spikes,
            inv_link_func,
        ) = coupled_model_simulate
        if delta_n_neuron != 0:
            feedforward_input = np.zeros(
                (
                    feedforward_input.shape[0],
                    feedforward_input.shape[1] + delta_n_neuron,
                    feedforward_input.shape[2],
                )
            )
        with expectation:
            simulation.simulate_recurrent(
                coupling_coeff,
                feedforward_coeff,
                intercepts,
                random_key,
                feedforward_input,
                coupling_basis,
                init_spikes,
                inv_link_func,
            )

    @pytest.mark.parametrize(
        "delta_dim, expectation",
        [
            (
                -1,
                pytest.raises(
                    ValueError, match="`feedforward_input` must be three-dimensional"
                ),
            ),
            (0, does_not_raise()),
            (
                1,
                pytest.raises(
                    ValueError, match="`feedforward_input` must be three-dimensional"
                ),
            ),
        ],
    )
    def test_simulate_input_dimensionality(
        self, delta_dim, expectation, coupled_model_simulate
    ):
        """
        Test the `simulate` method with input data of different dimensionalities.
        Ensure correct dimensionality for input.
        """
        (
            coupling_coeff,
            feedforward_coeff,
            intercepts,
            random_key,
            feedforward_input,
            coupling_basis,
            init_spikes,
            inv_link_func,
        ) = coupled_model_simulate
        if delta_dim == -1:
            feedforward_input = np.zeros(feedforward_input.shape[:2])
        elif delta_dim == 1:
            feedforward_input = np.zeros(feedforward_input.shape + (1,))
        with expectation:
            simulation.simulate_recurrent(
                coupling_coeff,
                feedforward_coeff,
                intercepts,
                random_key,
                feedforward_input,
                coupling_basis,
                init_spikes,
                inv_link_func,
            )

    @pytest.mark.parametrize(
        "delta_dim, expectation",
        [
            (0, does_not_raise()),
            (1, pytest.raises(ValueError, match="`init_y` must be two-dimensional")),
        ],
    )
    def test_simulate_y_dimensionality(
        self, delta_dim, expectation, coupled_model_simulate
    ):
        """
        Test the `simulate` method with init_spikes of different dimensionalities.
        Ensure correct dimensionality for init_spikes.
        """
        (
            coupling_coeff,
            feedforward_coeff,
            intercepts,
            random_key,
            feedforward_input,
            coupling_basis,
            init_spikes,
            inv_link_func,
        ) = coupled_model_simulate
        if delta_dim == -1:
            init_spikes = np.zeros((feedforward_input.shape[0],))
        elif delta_dim == 1:
            init_spikes = np.zeros(
                (feedforward_input.shape[0], feedforward_input.shape[1], 1)
            )
        with expectation:
            simulation.simulate_recurrent(
                coupling_coeff,
                feedforward_coeff,
                intercepts,
                random_key,
                feedforward_input,
                coupling_basis,
                init_spikes,
                inv_link_func,
            )

    @pytest.mark.parametrize(
        "delta_n_neuron, expectation",
        [
            (
                -1,
                pytest.raises(ValueError, match="The number of neurons"),
            ),
            (0, does_not_raise()),
            (
                1,
                pytest.raises(ValueError, match="The number of neurons"),
            ),
        ],
    )
    def test_simulate_n_neuron_match_y(
        self, delta_n_neuron, expectation, coupled_model_simulate
    ):
        """
        Test the `simulate` method to ensure that The number of neurons in init_spikes
        matches the model's parameters.
        """
        (
            coupling_coeff,
            feedforward_coeff,
            intercepts,
            random_key,
            feedforward_input,
            coupling_basis,
            init_spikes,
            inv_link_func,
        ) = coupled_model_simulate
        init_spikes = jnp.zeros(
            (init_spikes.shape[0], feedforward_input.shape[1] + delta_n_neuron)
        )
        with expectation:
            simulation.simulate_recurrent(
                coupling_coeff,
                feedforward_coeff,
                intercepts,
                random_key,
                feedforward_input,
                coupling_basis,
                init_spikes,
                inv_link_func,
            )

    @pytest.mark.parametrize(
        "delta_tp, expectation",
        [
            (
                -1,
                pytest.raises(ValueError, match="`init_y` and `coupling_basis_matrix`"),
            ),
            (0, does_not_raise()),
            (
                1,
                pytest.raises(ValueError, match="`init_y` and `coupling_basis_matrix`"),
            ),
        ],
    )
    def test_simulate_time_point_match_y(
        self, delta_tp, expectation, coupled_model_simulate
    ):
        """
        Test the `simulate` method to ensure that the time points in init_y
        are consistent with the coupling_basis window size (they must be equal).
        """
        (
            coupling_coeff,
            feedforward_coeff,
            intercepts,
            random_key,
            feedforward_input,
            coupling_basis,
            init_spikes,
            inv_link_func,
        ) = coupled_model_simulate
        init_spikes = jnp.zeros((init_spikes.shape[0] + delta_tp, init_spikes.shape[1]))
        with expectation:
            simulation.simulate_recurrent(
                coupling_coeff,
                feedforward_coeff,
                intercepts,
                random_key,
                feedforward_input,
                coupling_basis,
                init_spikes,
                inv_link_func,
            )

    @pytest.mark.parametrize(
        "delta_tp, expectation",
        [
            (
                -1,
                pytest.raises(ValueError, match="`init_y` and `coupling_basis_matrix`"),
            ),
            (0, does_not_raise()),
            (
                1,
                pytest.raises(ValueError, match="`init_y` and `coupling_basis_matrix`"),
            ),
        ],
    )
    def test_simulate_time_point_match_coupling_basis(
        self, delta_tp, expectation, coupled_model_simulate
    ):
        """
        Test the `simulate` method to ensure that the window size in coupling_basis
        is consistent with the time-points in init_spikes (they must be equal).
        """
        (
            coupling_coeff,
            feedforward_coeff,
            intercepts,
            random_key,
            feedforward_input,
            coupling_basis,
            init_spikes,
            inv_link_func,
        ) = coupled_model_simulate
        coupling_basis = jnp.zeros(
            (coupling_basis.shape[0] + delta_tp,) + coupling_basis.shape[1:]
        )
        with expectation:
            simulation.simulate_recurrent(
                coupling_coeff,
                feedforward_coeff,
                intercepts,
                random_key,
                feedforward_input,
                coupling_basis,
                init_spikes,
                inv_link_func,
            )

    @pytest.mark.parametrize(
        "delta_features, expectation",
        [
            (
                -1,
                pytest.raises(
                    ValueError,
                    match="Inconsistent number of features. spike basis coefficients has",
                ),
            ),
            (0, does_not_raise()),
            (
                1,
                pytest.raises(
                    ValueError,
                    match="Inconsistent number of features. spike basis coefficients has",
                ),
            ),
        ],
    )
    def test_simulate_feature_consistency_input(
        self, delta_features, expectation, coupled_model_simulate
    ):
        """
        Test the `simulate` method ensuring the number of features in `feedforward_input` is
        consistent with the model's expected number of features.

        Notes
        -----
        The total feature number `model.coef_.shape[1]` must be equal to
        `feedforward_input.shape[2] + coupling_basis.shape[1]*n_neurons`
        """
        (
            coupling_coeff,
            feedforward_coeff,
            intercepts,
            random_key,
            feedforward_input,
            coupling_basis,
            init_spikes,
            inv_link_func,
        ) = coupled_model_simulate
        feedforward_input = jnp.zeros(
            (
                feedforward_input.shape[0],
                feedforward_input.shape[1],
                feedforward_input.shape[2] + delta_features,
            )
        )
        with expectation:
            simulation.simulate_recurrent(
                coupling_coeff,
                feedforward_coeff,
                intercepts,
                random_key,
                feedforward_input,
                coupling_basis,
                init_spikes,
                inv_link_func,
            )

    @pytest.mark.parametrize(
        "delta_features, expectation",
        [
            (-1, pytest.raises(ValueError, match="Inconsistent number of features")),
            (0, does_not_raise()),
            (1, pytest.raises(ValueError, match="Inconsistent number of features")),
        ],
    )
    def test_simulate_feature_consistency_coupling_basis(
        self, delta_features, expectation, coupled_model_simulate
    ):
        """
        Test the `simulate` method ensuring the number of features in `coupling_basis` is
        consistent with the model's expected number of features.

        Notes
        -----
        The total feature number `model.coef_.shape[1]` must be equal to
        `feedforward_input.shape[2] + coupling_basis.shape[1]*n_neurons`
        """
        (
            coupling_coeff,
            feedforward_coeff,
            intercepts,
            random_key,
            feedforward_input,
            coupling_basis,
            init_spikes,
            inv_link_func,
        ) = coupled_model_simulate
        coupling_basis = jnp.zeros(
            (coupling_basis.shape[0], coupling_basis.shape[1] + delta_features)
        )
        with expectation:
            simulation.simulate_recurrent(
                coupling_coeff,
                feedforward_coeff,
                intercepts,
                random_key,
                feedforward_input,
                coupling_basis,
                init_spikes,
                inv_link_func,
            )



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp
from jax.test_util import check_grads

from nemos.third_party.jaxopt.jaxopt.tree_util import tree_l2_norm
from nemos.third_party.jaxopt.jaxopt import FixedPointIteration
from nemos.third_party.jaxopt.jaxopt._src import test_util



class FixedPointIterationTest(test_util.JaxoptTestCase):

  def test_geometric_decay(self):
    """Test convergence for geometric progression with common ratio r < 1."""
    def f(x):
      return jnp.array([0.9, 0.99, 0.999]) * x
    x0 = jnp.array([1., 0.4, -0.2])
    tol = 1e-5
    sol, state = FixedPointIteration(f, maxiter=8000, tol=tol).run(x0)
    self.assertLess(state.error, tol)
    sol_norm = tree_l2_norm(sol)
    big_tol = 1e-2  # we are forced to take high tol since convergence is very slow for 0.999
    self.assertLess(sol_norm, big_tol)

  @parameterized.product(jit=[False,True])
  def test_sin_fixed_point(self, jit):
    """Test convergence for simple polynomials and sin.
    Also test the support of pytree in input/output.
    """
    def f(x):  # Another fixed point exists for x[0] : ~1.11
      return jnp.sin(x[0]) * (x[0] ** 2), x[1] ** 3
    x0 = jnp.array([0.6, 0., -0.1]), jnp.array([[0.7], [0.5]])
    tol = 1e-6
    sol, state = FixedPointIteration(f, maxiter=100, tol=tol, jit=jit).run(x0)
    self.assertLess(state.error, tol)
    sol_norm = tree_l2_norm(sol)
    self.assertLess(sol_norm, tol)
    f_fixed_point = jnp.zeros(3), jnp.zeros((2,1))
    self.assertAllClose(sol, f_fixed_point)

  def test_cos_fixed_point(self):
    """Test convergence for cos fixed point (non-zero fixed point).
    Also test support for additional parameters.
    """
    def f(x, theta):
      return jnp.cos(x) + theta
    x0 = jnp.array([0.3, jnp.pi / 4])
    theta = jnp.array([0., jnp.pi / 2])
    f_fixed_point = jnp.array([0.73908614842288, jnp.pi / 2])
    tol = 1e-2
    sol, state = FixedPointIteration(f, maxiter=100000, tol=tol).run(x0, theta)
    self.assertLess(state.error, tol)
    self.assertArraysAllClose(sol, f_fixed_point, atol=1e-2)

  def test_has_aux(self):
    """Test support of ``has_aux`` functionnality."""
    def f(x, r, y):
      fx = jnp.cos(x)  # fixed point 0.739085
      fy = r * y * (1 - y)  # logistic map: chaotic behavior
      return fx, fy
    r = jnp.array([3.95])  # for this value there is chaotic behavior
    x, y = jnp.array([1.]), jnp.array([0.6])
    tol = 2e-2
    fp = FixedPointIteration(f, has_aux=True)
    state = fp.init_state(x, r, y)
    sols = []
    for i in range(10):
      x, state = fp.update(x, state, r, y)
      sols.append(x)
      y = state.aux
    self.assertLess(state.error, tol)
    self.assertArraysAllClose(x, jnp.array([0.739085]), atol=1e2)

  @parameterized.product(implicit_diff=[False, True])
  def test_simple_grads(self, implicit_diff):
    """Test correctness of gradients on a simple function."""
    def f(x, theta):
      return jnp.cos(x) + theta
    x0 = jnp.array([0., jnp.pi / 4])
    theta = jnp.array([0., jnp.pi / 2])
    fp = FixedPointIteration(f, maxiter=4*1000, tol=1e-6, implicit_diff=implicit_diff)
    def solve_run(args, kwargs):
      return fp.run(x0, *args, **kwargs)[0]
    check_grads(solve_run, args=([theta], {}), order=1, modes=['rev'], eps=1e-3)

  def test_grads_flat_landscape(self):
    """Test correctness of gradients."""

    def f(x, theta):
      return theta * x, x

    x0 = jnp.array([1.,-0.5])
    theta = jnp.array([0.05, 0.2])
    fp = FixedPointIteration(f, maxiter=1000, tol=1e-6, has_aux=True)

    def solve_run(theta):
      return fp.run(x0, theta).params[0]

    check_grads(solve_run, args=(theta,), order=1, modes=['rev'], eps=1e-4)

  def test_affine_contractive_mapping(self):
    """Test correctness on big affine contractive mapping."""
    n = 200
    key, subkey = jax.random.split(jax.random.PRNGKey(0))
    M = jax.random.uniform(subkey, shape=(n,n), minval=-1., maxval=1.)
    M = M + M.T  # random symmetric matrix
    eigv = jnp.linalg.eigvalsh(M)
    spectral_radius = jnp.abs(eigv[-1])
    eps = 1e-4
    M = M / (2*spectral_radius + eps)  # contractive mapping: lbda_max < 0.5
    key, subkey = jax.random.split(key)
    b = jax.random.uniform(subkey, shape=(n,))
    def f(x, M, b):
      return M @ x + b
    tol = 1e-6
    fp = FixedPointIteration(f, maxiter=100, tol=tol)
    x0 = jnp.zeros_like(b)
    sol, state = fp.run(x0, M, b)
    self.assertLess(state.error, tol)
    def solve_run(args, kwargs):
      return fp.run(x0, *args, **kwargs)[0]
    check_grads(solve_run, args=([M, b], {}), order=1, modes=['rev'], eps=1e-4)


if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  # jax.config.update("jax_enable_x64", True)
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common tests."""

from functools import partial
from absl.testing import absltest
from absl.testing import parameterized


from contextlib import redirect_stdout
import io


import jax
import jax.numpy as jnp
import optax
from sklearn import datasets
from sklearn import preprocessing

import nemos.third_party.jaxopt.jaxopt as jaxopt
from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt import projection
from nemos.third_party.jaxopt.jaxopt import prox
from nemos.third_party.jaxopt.jaxopt import tree_util as tu
from nemos.third_party.jaxopt.jaxopt._src import test_util
from nemos.third_party.jaxopt.jaxopt._src.base import LinearOperator
from nemos.third_party.jaxopt.jaxopt._src.objective import (
    LeastSquares,
    MulticlassLinearSvmDual,
)
from nemos.third_party.jaxopt.jaxopt._src.osqp import BoxOSQP


N_CALLS = 0


def check_states_have_same_types(state1, state2):
    if len(state1._fields) != len(state2._fields):
        raise ValueError(
            "state1 and state2 should have the same number of " "attributes."
        )

    for attr1, attr2 in zip(state1._fields, state2._fields):
        if attr1 != attr2:
            raise ValueError(
                "Attribute names do not agree: %s and %s." % (attr1, attr2)
            )

        type1 = type(getattr(state1, attr1)).__name__
        type2 = type(getattr(state2, attr2)).__name__

        if type1 != type2:
            raise ValueError(
                "Attribute '%s' has different types in state1 and "
                "state2: %s vs. %s" % (attr1, type1, type2)
            )


class CommonTest(test_util.JaxoptTestCase):
    def test_jit_update_and_returned_states(self):
        fun = objective.least_squares
        root_fun = jax.grad(fun)

        def fixed_point_fun(params, *args, **kwargs):
            return root_fun(params, *args, **kwargs) - params

        X, y = datasets.make_classification(
            n_samples=10, n_features=5, n_classes=3, n_informative=3, random_state=0
        )
        data = (X, y)
        params0 = jnp.zeros(X.shape[1])

        fpi = jaxopt.FixedPointIteration(fixed_point_fun=fixed_point_fun)

        # Gradient solvers and fixed point
        for solver in (
            jaxopt.GradientDescent(fun=fun),
            jaxopt.PolyakSGD(fun=fun),
            jaxopt.OptaxSolver(opt=optax.adam(1e-3), fun=fun),
            fpi,
            jaxopt.Broyden(fun=root_fun),
            jaxopt.AndersonAcceleration(fixed_point_fun=fixed_point_fun),
            jaxopt.AndersonWrapper(fpi),
            jaxopt.ArmijoSGD(fun=fun),
            jaxopt.BFGS(fun),
            jaxopt.LBFGS(fun=fun),
        ):
            hash(solver)  # Checks that hash works.
            update = jax.jit(solver.update)
            state0 = solver.init_state(params0, data=data)
            params, state = update(params0, state0, data=data)
            check_states_have_same_types(state0, state)

        # Proximal gradient solvers.
        for solver in (
            jaxopt.ProximalGradient(fun=fun, prox=prox.prox_lasso),
            jaxopt.BlockCoordinateDescent(fun=fun, block_prox=prox.prox_lasso),
        ):
            hash(solver)  # Checks that hash works.
            update = jax.jit(solver.update)
            state0 = solver.init_state(params0, hyperparams_prox=1.0, data=data)
            params, state = update(params0, state0, hyperparams_prox=1.0, data=data)
            check_states_have_same_types(state0, state)

        # Projected gradient solvers.
        l2_proj = projection.projection_simplex

        for solver in (jaxopt.ProjectedGradient(fun=fun, projection=l2_proj),):
            hash(solver)  # Checks that hash works.
            update = jax.jit(solver.update)
            state0 = solver.init_state(params0, hyperparams_proj=1.0, data=data)
            params, state = update(params0, state0, hyperparams_proj=1.0, data=data)
            check_states_have_same_types(state0, state)

        # Bisection
        optimality_fun = lambda x: x**3 - x - 2
        bisec = jaxopt.Bisection(optimality_fun=optimality_fun, lower=1, upper=2)

        hash(bisec)  # Checks that hash works.
        update = jax.jit(bisec.update)
        state0 = bisec.init_state()
        params, state = update(params=None, state=state0)
        check_states_have_same_types(state0, state)

    def test_n_calls(self):
        global N_CALLS

        class LeastSquaresNCalls(LeastSquares):
            def __init__(self, with_custom_linop=False):
                super().__init__()
                self.with_custom_linop = with_custom_linop

            def __call__(self, W, data):
                global N_CALLS
                N_CALLS += 1
                return super().__call__(W, data)

            def make_linop(self, data):
                """Creates linear operator."""
                if self.with_custom_linop:
                    return LinopNCall(data[0])
                else:
                    return super().make_linop(data)

        class LinopNCall(LinearOperator):
            def matvec(self, x):
                """Computes dot(A, x)."""
                global N_CALLS
                N_CALLS += 1
                return jnp.dot(self.A, x)

            def rmatvec_element(self, x, idx):
                """Computes dot(A.T, x)[idx]."""
                global N_CALLS
                N_CALLS += 1
                return jnp.dot(self.A[:, idx], x)

        fun = LeastSquaresNCalls()
        root_fun = jax.grad(fun)

        def fixed_point_fun(params, *args, **kwargs):
            return root_fun(params, *args, **kwargs) - params

        X, y = datasets.make_classification(
            n_samples=10, n_features=5, n_classes=3, n_informative=3, random_state=0
        )
        data = (X, y)
        params0 = jnp.zeros(X.shape[1])
        common_kwargs = dict(jit=False, maxiter=30)
        fpi = jaxopt.FixedPointIteration(
            fixed_point_fun=fixed_point_fun, **common_kwargs
        )

        # Gradient solvers and fixed point
        for solver in (
            jaxopt.PolyakSGD(fun=fun, **common_kwargs),
            fpi,
            jaxopt.Broyden(fun=root_fun, **common_kwargs),
            jaxopt.AndersonAcceleration(
                fixed_point_fun=fixed_point_fun, **common_kwargs
            ),
            jaxopt.NonlinearCG(fun, **common_kwargs),
            jaxopt.BFGS(fun, **common_kwargs),
            jaxopt.LBFGS(fun=fun, **common_kwargs),
        ):
            _, state = solver.run(params0, data=data)
            self.assertEqual(state.num_fun_eval, N_CALLS)
            N_CALLS = 0

        # # Proximal gradient solvers.
        # FIXME: currently the following test does not work because
        # of a fori_loop issue in block cd (does not allow non-jitted functions). (zramzi)
        # fun = LeastSquaresNCalls(with_custom_linop=True)
        # for solver in (jaxopt.BlockCoordinateDescent(fun=fun,
        #                                              block_prox=prox.prox_lasso,
        #                                              **common_kwargs),):

        #   _, state = solver.run(params0, hyperparams_prox=1.0, data=data)
        #   self.assertEqual(state.num_fun_eval, N_CALLS)
        #   N_CALLS = 0

        # Mirror Descent
        Y = preprocessing.LabelBinarizer().fit_transform(y)
        Y = jnp.asarray(Y)
        n_samples, n_classes = Y.shape

        class MulticlassLinearSvmDualNCall(MulticlassLinearSvmDual):
            def __call__(self, params, *args, **kwargs):
                global N_CALLS
                N_CALLS += 1
                return super().__call__(params, *args, **kwargs)

        fun = MulticlassLinearSvmDualNCall()
        lam = 10.0
        data = (X, Y)

        beta_init = jnp.ones((n_samples, n_classes)) / n_classes

        def kl_projection(x, hyperparams_proj):
            del hyperparams_proj
            return tu.tree_map(lambda t: jax.nn.softmax(t, -1), x)

        # Generating function of the Bregman divergence.
        kl_generating_fun = lambda x: -jnp.sum(jax.scipy.special.entr(x) + x)
        # Row-wise mirror map.
        kl_mapping_fun = jax.vmap(jax.grad(kl_generating_fun))
        projection_grad = jaxopt.MirrorDescent.make_projection_grad(
            kl_projection, kl_mapping_fun
        )
        md = jaxopt.MirrorDescent(
            fun=fun, projection_grad=projection_grad, stepsize=1e-3, **common_kwargs
        )
        _, state = md.run(beta_init, None, lam, data)
        self.assertEqual(state.num_fun_eval, N_CALLS)
        N_CALLS = 0

        # Bisection
        def opt_fun(x):
            global N_CALLS
            N_CALLS += 1
            return x**3 - x - 2

        bisec = jaxopt.Bisection(
            optimality_fun=opt_fun,
            lower=1,
            upper=2,
            **common_kwargs,
        )

        _, state = bisec.run()
        self.assertEqual(state.num_fun_eval, N_CALLS)
        N_CALLS = 0

    @parameterized.product(value_and_grad=[True, False])
    def test_aux_consistency(self, value_and_grad):
        def fun_(w, X, y):
            residuals = jnp.dot(X, w) - y
            return jnp.sum(residuals**2), residuals

        if value_and_grad:
            fun = jax.value_and_grad(fun_, has_aux=True)
        else:
            fun = fun_
        X, y = datasets.make_classification(
            n_samples=10, n_features=5, n_classes=3, n_informative=3, random_state=0
        )

        for solver in (
            jaxopt.OptaxSolver(
                opt=optax.adam(1e-3),
                fun=fun,
                has_aux=True,
                value_and_grad=value_and_grad,
            ),
            jaxopt.BFGS(fun=fun, has_aux=True, value_and_grad=value_and_grad),
            jaxopt.LBFGS(fun=fun, has_aux=True, value_and_grad=value_and_grad),
            jaxopt.NonlinearCG(fun=fun, has_aux=True, value_and_grad=value_and_grad),
            jaxopt.GradientDescent(
                fun=fun, has_aux=True, value_and_grad=value_and_grad
            ),
            jaxopt.ArmijoSGD(fun=fun, has_aux=True, value_and_grad=value_and_grad),
            jaxopt.PolyakSGD(fun=fun, has_aux=True, value_and_grad=value_and_grad),
        ):
            params = jnp.zeros(X.shape[1])
            state1 = solver.init_state(params, X, y)
            params, state2 = solver.update(params, state1, X, y)
            self.assertEqual(type(state1.aux), type(state2.aux))

    def test_dtype_consistency(self):
        # We don't use float32 (default dtype in JAX)
        # to detect when the dtype is incorrectly inherited.
        # We use two different dtypes for function outputs and
        # for parameters, as it's a common use case.
        dtype_fun = jnp.bfloat16
        dtype_params = jnp.float16

        def fun(w, X, y):
            w = w.astype(dtype_fun)
            return jnp.sum((jnp.dot(X, w) - y) ** 2)

        X, y = datasets.make_classification(
            n_samples=10, n_features=5, n_classes=3, n_informative=3, random_state=0
        )
        X = jnp.array(X, dtype=dtype_fun)
        y = jnp.array(y, dtype=dtype_fun)
        init = jnp.zeros(X.shape[1], dtype=dtype_params)

        for solver in (
            jaxopt.ArmijoSGD(fun=fun),
            jaxopt.BFGS(fun, linesearch="zoom"),
            jaxopt.BFGS(fun, linesearch="backtracking"),
            jaxopt.GradientDescent(fun=fun, acceleration=True),
            jaxopt.GradientDescent(fun=fun, acceleration=False),
            jaxopt.GradientDescent(fun=fun, stepsize=1e-3),
            jaxopt.LBFGS(fun=fun, linesearch="zoom"),
            jaxopt.LBFGS(fun=fun, linesearch="backtracking"),
            jaxopt.NonlinearCG(fun, linesearch="zoom"),
            jaxopt.NonlinearCG(fun, linesearch="backtracking"),
            jaxopt.OptaxSolver(opt=optax.adam(1e-3), fun=fun),
            jaxopt.PolyakSGD(fun=fun, momentum=0.0),
            jaxopt.PolyakSGD(fun=fun, momentum=0.9),
            jaxopt.ProjectedGradient(
                fun=fun, projection=jaxopt.projection.projection_non_negative
            ),
        ):
            state = solver.init_state(init, X=X, y=y)
            params, state = solver.update(init, state, X=X, y=y)

            self.assertEqual(params.dtype, dtype_params)

            solver_name = solver.__class__.__name__

            if hasattr(state, "value"):
                # FIXME: ideally, solver states should always include a value attribute.
                self.assertEqual(
                    state.value.dtype,
                    dtype_fun,
                    msg="value dtype error in solver '%s'" % solver_name,
                )

            # Usually, the error is computed from the gradient,
            # which has the same dtype as the parameters.
            self.assertEqual(state.error.dtype, dtype_params)

    def test_weak_type_consistency(self):
        def fun(w, X, y):
            return jnp.sum((jnp.dot(X, w) - y) ** 2)

        solvers = (
            jaxopt.ArmijoSGD(fun=fun),
            jaxopt.BFGS(fun, linesearch="zoom"),
            jaxopt.BFGS(fun, linesearch="backtracking"),
            jaxopt.GradientDescent(fun=fun, acceleration=True),
            jaxopt.GradientDescent(fun=fun, acceleration=False),
            jaxopt.LBFGS(fun=fun, linesearch="zoom"),
            jaxopt.LBFGS(fun=fun, linesearch="backtracking"),
            jaxopt.NonlinearCG(fun, linesearch="zoom"),
            jaxopt.NonlinearCG(fun, linesearch="backtracking"),
            jaxopt.OptaxSolver(opt=optax.adam(1e-3), fun=fun),
            jaxopt.PolyakSGD(fun=fun, momentum=0.0),
            jaxopt.PolyakSGD(fun=fun, momentum=0.9),
            jaxopt.ProjectedGradient(
                fun=fun, projection=jaxopt.projection.projection_non_negative
            ),
        )

        X, y = datasets.make_classification(
            n_samples=10, n_features=5, n_classes=3, n_informative=3, random_state=0
        )
        X = jnp.array(X)
        y = jnp.array(y)
        init = jnp.zeros(X.shape[1])

        for solver in solvers:
            state0 = solver.init_state(init, X=X, y=y)
            params, state1 = solver.update(init, state0, X=X, y=y)
            for field in state0._fields:
                field0 = getattr(state0, field)
                field1 = getattr(state1, field)

                weak_type0 = getattr(field0, "weak_type", None)
                weak_type1 = getattr(field1, "weak_type", None)

                solver_name = solver.__class__.__name__
                msg = "weak_type inconsistency for attribute '%s' in solver '%s'"
                self.assertEqual(weak_type0, weak_type1, msg=msg % (field, solver_name))

    @parameterized.product(verbose=[True, False])
    def test_jit_with_or_without_verbose(self, verbose):
        fun = lambda p: p @ p

        def fixed_point_fun(params):
            return fun(params) - params

        solvers = (
            # Unconstrained
            jaxopt.GradientDescent(fun=fun, jit=True, verbose=verbose, maxiter=4),
            jaxopt.PolyakSGD(fun=fun, jit=True, verbose=verbose, maxiter=4),
            jaxopt.Broyden(fun=fixed_point_fun, jit=True, verbose=verbose, maxiter=4),
            jaxopt.AndersonAcceleration(
                fixed_point_fun=fixed_point_fun, jit=True, verbose=verbose, maxiter=4
            ),
            jaxopt.ArmijoSGD(fun=fun, jit=True, verbose=verbose, maxiter=4),
            jaxopt.BFGS(fun, linesearch="zoom", jit=True, verbose=verbose, maxiter=4),
            jaxopt.BFGS(
                fun, linesearch="backtracking", jit=True, verbose=verbose, maxiter=4
            ),
            jaxopt.BFGS(
                fun, linesearch="hager-zhang", jit=True, verbose=verbose, maxiter=4
            ),
            jaxopt.LBFGS(fun=fun, jit=True, verbose=verbose, maxiter=4),
            jaxopt.ArmijoSGD(fun=fun, jit=True, verbose=verbose, maxiter=4),
            jaxopt.NonlinearCG(fun, jit=True, verbose=verbose, maxiter=4),
            # Unconstrained, nonlinear least-squares
            jaxopt.GaussNewton(residual_fun=fun, jit=True, verbose=verbose, maxiter=4),
            jaxopt.LevenbergMarquardt(
                residual_fun=fun, jit=True, verbose=verbose, maxiter=4
            ),
            # Constrained
            jaxopt.ProjectedGradient(
                fun=fun,
                projection=jaxopt.projection.projection_non_negative,
                jit=True,
                verbose=verbose,
                maxiter=4,
            ),
            # Optax wrapper
            jaxopt.OptaxSolver(
                opt=optax.adam(1e-1), fun=fun, jit=True, verbose=verbose, maxiter=4
            ),
        )

        @partial(jax.jit, static_argnums=(1,))
        def run_solver(p0, solver):
            return solver.run(p0)

        for solver in solvers:
            stdout = io.StringIO()
            with redirect_stdout(stdout):
                jax.block_until_ready(run_solver(jnp.arange(2.0), solver))
            printed = len(stdout.getvalue()) > 0
            if verbose:
                self.assertTrue(printed)
            else:
                self.assertFalse(printed)

        # Proximal gradient solvers
        fun = objective.least_squares
        X, y = datasets.make_classification(
            n_samples=10, n_features=5, n_classes=3, n_informative=3, random_state=0
        )
        data = (X, y)
        params0 = jnp.zeros(X.shape[1])

        @partial(jax.jit, static_argnums=(1,))
        def run_solver_prox(p0, solver):
            return solver.run(p0, hyperparams_prox=1.0, data=data)

        for solver in (
            jaxopt.ProximalGradient(
                fun=fun, prox=prox.prox_lasso, jit=True, verbose=verbose, maxiter=4
            ),
            jaxopt.BlockCoordinateDescent(
                fun=fun,
                block_prox=prox.prox_lasso,
                jit=True,
                verbose=verbose,
                maxiter=4,
            ),
        ):
            stdout = io.StringIO()
            with redirect_stdout(stdout):
                jax.block_until_ready(run_solver_prox(params0, solver))
            printed = len(stdout.getvalue()) > 0
            if verbose:
                self.assertTrue(printed)
            else:
                self.assertFalse(printed)

        # Mirror Descent
        Y = preprocessing.LabelBinarizer().fit_transform(y)
        Y = jnp.asarray(Y)
        n_samples, n_classes = Y.shape

        fun = MulticlassLinearSvmDual()
        lam = 10.0
        data = (X, Y)

        beta_init = jnp.ones((n_samples, n_classes)) / n_classes

        def kl_projection(x, hyperparams_proj):
            del hyperparams_proj
            return tu.tree_map(lambda t: jax.nn.softmax(t, -1), x)

        # Generating function of the Bregman divergence.
        kl_generating_fun = lambda x: -jnp.sum(jax.scipy.special.entr(x) + x)
        # Row-wise mirror map.
        kl_mapping_fun = jax.vmap(jax.grad(kl_generating_fun))
        projection_grad = jaxopt.MirrorDescent.make_projection_grad(
            kl_projection, kl_mapping_fun
        )

        @jax.jit
        def run_mirror_descent(b0):
            md = jaxopt.MirrorDescent(
                fun=fun,
                projection_grad=projection_grad,
                stepsize=1e-3,
                maxiter=4,
                jit=True,
                verbose=verbose,
            )
            _, state = md.run(b0, None, lam, data)
            return state

        stdout = io.StringIO()
        with redirect_stdout(stdout):
            jax.block_until_ready(run_mirror_descent(beta_init))
        printed = len(stdout.getvalue()) > 0
        if verbose:
            self.assertTrue(printed)
        else:
            self.assertFalse(printed)

        # Quadratic programming - BoxOSQP
        x = jnp.array([1.0, 2.0])
        a = jnp.array([-0.5, 1.5])
        b = 0.3
        q = -x
        # Find ||y-x||^2 such that jnp.dot(y, a) = b.

        matvec_Q = lambda params_Q, u: u
        matvec_A = lambda params_A, u: jnp.dot(a, u).reshape(1)

        tol = 1e-4

        @jax.jit
        def run_box_osqp(params_obj, params_ineq):
            osqp = BoxOSQP(
                matvec_Q=matvec_Q,
                matvec_A=matvec_A,
                tol=tol,
                jit=True,
                verbose=verbose,
                maxiter=4,
            )
            return osqp.run(None, (None, params_obj), None, (params_ineq, params_ineq))

        stdout = io.StringIO()
        with redirect_stdout(stdout):
            jax.block_until_ready(run_box_osqp(q, b))
        printed = len(stdout.getvalue()) > 0
        if verbose:
            self.assertTrue(printed)
        else:
            self.assertFalse(printed)


if __name__ == "__main__":
    absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CVXPY wrappers."""

from typing import Callable
from typing import Optional

from dataclasses import dataclass

import jax.numpy as jnp 
from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt._src import implicit_diff as idf
from nemos.third_party.jaxopt.jaxopt._src import tree_util


def _check_params(params_obj, params_eq=None, params_ineq=None):
  if params_obj is None:
    raise ValueError("params_obj should be a tuple (Q, c)")
  
  Q, c = params_obj
  if Q.shape[0] != Q.shape[1]:
    raise ValueError("Q must be a square matrix.")
  if Q.shape[1] != c.shape[0]:
    raise ValueError("Q.shape[1] != c.shape[0]")

  if params_eq is not None:
    A, b = params_eq
    if A.shape[0] != b.shape[0]:
      raise ValueError("A.shape[0] != b.shape[0]")
    if A.shape[1] != Q.shape[1]:
      raise ValueError("Q.shape[1] != A.shape[1]")

  if params_ineq is not None:
    G, h = params_ineq
    if G.shape[0] != h.shape[0]:
      raise ValueError("G.shape[0] != h.shape[0]")
    if G.shape[1] != Q.shape[1]:
      raise ValueError("G.shape[1] != Q.shape[1]")


def _make_cvxpy_qp_optimality_fun():
  """Makes the optimality function for CVXPY quadratic programming.

  Returns:
    optimality_fun(params, params_obj, params_eq, params_ineq) where
      params = (primal_var, eq_dual_var, ineq_dual_var)
      params_obj = (Q, c)
      params_eq = (A, b) or None
      params_ineq = (G, h) or None
  """
  def obj_fun(primal_var, params_obj):
    Q, c = params_obj
    return 0.5 * jnp.vdot(primal_var, jnp.dot(Q, primal_var)) + jnp.vdot(primal_var, c)

  def eq_fun(primal_var, params_eq):
    if params_eq is None:
      return None
    A, b = params_eq
    return jnp.dot(A, primal_var) - b

  def ineq_fun(primal_var, params_ineq):
    if params_ineq is None:
      return None
    G, h = params_ineq
    return jnp.dot(G, primal_var) - h

  return idf.make_kkt_optimality_fun(obj_fun, eq_fun, ineq_fun)


@dataclass(eq=False)
class CvxpyQP(base.Solver):
  """Wraps CVXPY's quadratic solver with implicit diff support.

  No support for matvec, pytrees, jit and vmap.
  Meant to be run on CPU. Provide high precision solutions.

  The objective function is::

    0.5 * x^T Q x + c^T x subject to Gx <= h, Ax = b.
  
  Attributes:
    solver: string specifying the underlying solver used by Cvxpy, in ``"OSQP", "ECOS", "SCS"`` (default: ``"OSQP"``).
  """
  solver: str = 'OSQP'  #TODO(lbethune): "True" original OSQP implementation (not the one written in Jax). Confusing for user ?
  implicit_diff_solve: Optional[Callable] = None

  def run(self,
          init_params: Optional[jnp.ndarray],  # unused
          params_obj: base.ArrayPair,
          params_eq: Optional[base.ArrayPair] = None,
          params_ineq: Optional[base.ArrayPair] = None) -> base.OptStep:
    """Runs the quadratic programming solver in CVXPY.

    The returned params contains both the primal and dual solutions.

    Args:
      init_params: ignored.
      params_obj: (Q, c).
      params_eq: (A, b) or None if no equality constraints.
      params_ineq: (G, h) or None if no inequality constraints.
    Returns:
      (params, state), ``params = (primal_var, dual_var_eq, dual_var_ineq)``
    """
    # TODO(frostig,mblondel): experiment with `jax.experimental.host_callback`
    # to "support" other devices (GPU/TPU) in the interim, by calling into the
    # host CPU and running cvxpy there.
    #
    # TODO(lbethune): the interface of CVXPY could easily allow pytrees for constraints,
    # by populating `constraints` list with different Ai x = bi and Gi x <= hi.
    # Pytree support for x could be possible by creating a cp.Variable for each leaf in the pytree c.
    import cvxpy as cp

    del init_params  # no warm start
    _check_params(params_obj, params_eq, params_ineq)

    Q, c = params_obj
    x = cp.Variable(len(c))
    objective = 0.5 * cp.quad_form(x, Q) + c.T @ x

    constraints = []
    if params_eq is not None:
      A, b = params_eq
      constraints.append(A @ x == b)
    if params_ineq is not None:
      G, h = params_ineq
      constraints.append(G @ x <= h)

    pb = cp.Problem(cp.Minimize(objective), constraints)
    pb.solve(solver=self.solver)

    if pb.status in ["infeasible", "unbounded"]:
      raise ValueError("The problem is %s." % pb.status)

    dual_eq = None if params_eq is None else jnp.array(pb.constraints[0].dual_value)
    dual_ineq = None if params_ineq is None else jnp.array(pb.constraints[-1].dual_value)

    sol = base.KKTSolution(primal=jnp.array(x.value),
                           dual_eq=dual_eq,
                           dual_ineq=dual_ineq)

    # TODO(lbethune): pb.solver_stats is a "state" the user might be interested in.
    return base.OptStep(params=sol, state=None)

  def l2_optimality_error(
      self,
      params: jnp.ndarray,
      params_obj: base.ArrayPair,
      params_eq: Optional[base.ArrayPair],
      params_ineq: Optional[base.ArrayPair],
  ) -> base.OptStep:
    """Computes the L2 norm of the KKT residuals."""
    pytree = self.optimality_fun(params, params_obj, params_eq, params_ineq)
    return tree_util.tree_l2_norm(pytree)

  def __post_init__(self):
    self.optimality_fun = _make_cvxpy_qp_optimality_fun()

    # Set up implicit diff.
    decorator = idf.custom_root(self.optimality_fun, has_aux=True,
                                solve=self.implicit_diff_solve)
    # pylint: disable=g-missing-from-attributes
    self.run = decorator(self.run)



"""Abstract class for estimators."""

# required to get ArrayLike to render correctly
from __future__ import annotations

import warnings
from collections import defaultdict
from typing import Any

from sklearn.base import BaseEstimator


class Base(BaseEstimator):
    """Base class for NeMoS estimators.

    A base class for estimators with utilities for getting and setting parameters,
    and for interacting with specific devices like CPU, GPU, and TPU.

    This class provides utilities for:
    - Getting and setting parameters using introspection.
    - Sending arrays to target devices (CPU, GPU, TPU).

    Notes
    -----
    The class provides helper methods mimicking scikit-learn's get_params and set_params.
    Additionally, it has methods for selecting target devices and sending arrays to them.
    """

    def get_params(self, deep=True) -> dict:
        """
        From scikit-learn, get parameters by inspecting init.

        Parameters
        ----------
        deep:
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        :
            A dictionary containing the parameters. Key is the parameter
            name, value is the parameter value.
        """
        return super().get_params(deep=deep)

    def set_params(self, **params: Any):
        """Set the parameters of this estimator.

        The method works on simple estimators as well as on nested objects
        (such as :class:`~sklearn.pipeline.Pipeline`). The latter have
        parameters of the form ``<component>__<parameter>`` so that it's
        possible to update each component of a nested object.

        Parameters
        ----------
        **params : dict
            Estimator parameters.

        Returns
        -------
        self : estimator instance
            Estimator instance.
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        get_params = getattr(self, "__sklearn_get_params__", self.get_params)
        valid_params = get_params(deep=True)
        nested_params: defaultdict = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition("__")
            if key not in valid_params:
                local_valid_params = self._get_param_names()
                raise ValueError(
                    f"Invalid parameter {key!r} for estimator {self}. "
                    f"Valid parameters are: {local_valid_params!r}."
                )

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value

        for key, sub_params in nested_params.items():
            # TODO(1.4): remove specific handling of "base_estimator".
            # The "base_estimator" key is special. It was deprecated and
            # renamed to "estimator" for several estimators. This means we
            # need to translate it here and set sub-parameters on "estimator",
            # but only if the user did not explicitly set a value for
            # "base_estimator".
            if (
                key == "base_estimator"
                and valid_params[key] == "deprecated"
                and self.__module__.startswith("sklearn.")
            ):
                warnings.warn(
                    (
                        f"Parameter 'base_estimator' of {self.__class__.__name__} is"
                        " deprecated in favor of 'estimator'. See"
                        f" {self.__class__.__name__}'s docstring for more details."
                    ),
                    FutureWarning,
                    stacklevel=2,
                )
                key = "estimator"
            valid_params[key].set_params(**sub_params)

        return self



"""Custom solvers module."""

from ._jaxopt_solvers import (
    JaxoptBFGS,
    JaxoptGradientDescent,
    JaxoptLBFGS,
    JaxoptNonlinearCG,
    JaxoptProximalGradient,
)
from ._optax_optimistix_solvers import (
    OptimistixOptaxGradientDescent,
    OptimistixOptaxLBFGS,
    OptimistixOptaxProximalGradient,
)
from ._optimistix_solvers import OptimistixBFGS, OptimistixNonlinearCG
from ._solver_doc_helper import get_solver_documentation
from ._solver_registry import list_available_solvers, solver_registry
from ._svrg import SVRG, ProxSVRG, WrappedProxSVRG, WrappedSVRG
from ._svrg_defaults import (
    glm_softplus_poisson_l_max_and_l,
    svrg_optimal_batch_and_stepsize,
)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt._src import test_util

import numpy as onp

from typing import Any
from typing import Callable
from typing import NamedTuple

import dataclasses

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt._src import base


class DummySolverState(NamedTuple):
  iter_num: int
  error: float
  value: float
  aux: Any


@dataclasses.dataclass(eq=False)
class DummySolver(base.IterativeSolver):
  """Dummy solver."""

  fun: Callable
  maxiter: int = 500
  tol: float = 1e-3
  implicit_diff: bool = False
  implicit_diff_solve = None
  jit = True
  unroll = False
  verbose = True

  def init_state(self, init_params: Any, *args, **kwargs) -> DummySolverState:
    return DummySolverState(iter_num=0, error=jnp.inf, value=jnp.inf, aux=None)

  def update(self,
             params: Any,
             state: DummySolverState,
             *args,
             **kwargs) -> base.OptStep:
    return base.OptStep(params=params, state=state)

  def dummy_method(self):
    return self

  def optimality_fun(self, *args, **kwargs):
    return jax.grad(self.fun)(*args, **kwargs)

  def __post_init__(self):
    self.dummy_attr = True


class BaseTest(test_util.JaxoptTestCase):

  def test_linear_operator(self):
    rng = onp.random.RandomState(0)
    A = rng.randn(5, 3)
    x = rng.randn(3)
    y = rng.randn(5)
    I_x = jnp.eye(3)
    I_y = jnp.eye(5)
    delta_x = rng.randn(1)[0]
    delta_y = rng.randn(1)[0]
    X = rng.randn(3, 2)
    delta_X = rng.randn(2)
    Y = rng.randn(5, 2)
    delta_Y = rng.randn(5)
    linop = base.LinearOperator(A)

    # Check matrix-vector operations.
    Ax = jnp.dot(A, x)
    self.assertArraysAllClose(linop.matvec(x), Ax)
    ATy = jnp.dot(A.T, y)
    self.assertArraysAllClose(linop.rmatvec(y), ATy)

    for i in range(A.shape[0]):
      self.assertAllClose(linop.matvec_element(x, i), Ax[i])
      self.assertArraysAllClose(linop.update_rmatvec(ATy, delta_y, i),
                                jnp.dot(A.T, y + delta_y * I_y[i]))

    for j in range(A.shape[1]):
      self.assertAllClose(linop.rmatvec_element(y, j), ATy[j])
      self.assertArraysAllClose(linop.update_matvec(Ax, delta_x, j),
                                jnp.dot(A, x + delta_x * I_x[j]))

    # Check matrix-matrix operations.
    def E(i, shape):
      ret = onp.zeros(shape)
      ret[i] = 1
      return ret

    AX = jnp.dot(A, X)
    self.assertArraysAllClose(linop.matvec(X), AX)
    ATY = jnp.dot(A.T, Y)

    self.assertArraysAllClose(linop.rmatvec(Y), ATY)
    for i in range(A.shape[0]):
      self.assertAllClose(linop.matvec_element(X, i), AX[i])
      # todo: implement this
      # self.assertArraysAllClose(linop.update_rmatvec(ATY, delta_Y, i),
      #                    jnp.dot(A.T, Y + delta_Y[:, None] * E(i, Y.shape)))

    for j in range(A.shape[1]):
      self.assertAllClose(linop.rmatvec_element(Y, j), ATY[j])
      self.assertArraysAllClose(linop.update_matvec(AX, delta_X, j),
                                jnp.dot(A, X + delta_X * E(j, X.shape)))

    # Check that flatten and unflatten work.
    leaf_values, treedef = jax.tree_util.tree_flatten(linop)
    linop2 = jax.tree_util.tree_unflatten(treedef, leaf_values)
    self.assertArraysAllClose(linop2.matvec(x), Ax)

  def test_solver_attributes(self):
    fun = lambda x: x
    solver = DummySolver(fun=fun, maxiter=10, tol=1.0, implicit_diff=True)
    self.assertEqual(solver.attribute_names(),
                     ("fun", "maxiter", "tol", "implicit_diff"))
    self.assertEqual(solver.attribute_values(), (fun, 10, 1.0, True))

  def test_solver_hash(self):
    fun = lambda x: x
    solver = DummySolver(fun=fun, maxiter=10, tol=1.0, implicit_diff=True)
    hash(solver)

  def test_solver_equality(self):
    fun = lambda x: x
    solver = DummySolver(fun=fun, maxiter=10, tol=1.0, implicit_diff=True)
    self.assertTrue(solver == solver)

  def test_jit_update(self):
    fun = lambda x: x
    solver = DummySolver(fun=fun, maxiter=10, tol=1.0, implicit_diff=True)
    update = jax.jit(solver.update)

  def test_maxiter_static_vs_dynamic(self):
    def f(maxiter):
      solver = DummySolver(
          fun=lambda x: x, maxiter=maxiter, tol=1.0, implicit_diff=True)
      return solver.run(7.0)

    f_jit_d = jax.jit(f)
    f_jit_s = jax.jit(f, static_argnums=0)

    ref0 = f(0)
    ref3 = f(3)
    self.assertAllClose(ref0, f_jit_s(0))
    self.assertAllClose(ref3, f_jit_s(3))
    self.assertAllClose(ref0, f_jit_d(0))
    self.assertAllClose(ref3, f_jit_d(3))

  def _assert_items_equal(self, items):
    """Asserts that all items are equal."""
    for i in range(1, len(items)):
      self.assertAllClose(items[0], items[i], err_msg=f"items[0] != items[{i}]")

  def test_make_funs_with_aux(self):
    def fun_with_aux(w, a, b):
      res = w * a - b
      return res ** 2, res

    def fun(w, a, b):
      return fun_with_aux(w, a, b)[0]

    vg_with_aux = jax.value_and_grad(fun_with_aux, has_aux=True)
    vg = jax.value_and_grad(fun)

    fun1, grad1, vg1 = base._make_funs_with_aux(fun_with_aux,
                                                value_and_grad=False,
                                                has_aux=True)
    fun2, grad2, vg2 = base._make_funs_with_aux(fun,
                                                value_and_grad=False,
                                                has_aux=False)
    fun3, grad3, vg3 = base._make_funs_with_aux(
        vg_with_aux, value_and_grad=True, has_aux=True
    )
    fun4, grad4, vg4 = base._make_funs_with_aux(
        vg, value_and_grad=True, has_aux=False
    )
    fun5, grad5, vg5 = base._make_funs_with_aux(
        fun_with_aux, value_and_grad=vg_with_aux, has_aux=True
    )
    fun6, grad6, vg6 = base._make_funs_with_aux(
        fun, value_and_grad=vg, has_aux=False
    )

    w, a, b = jnp.array([1.0, 2.0, 3.0])

    v1, aux1 = fun1(w, a, b)
    v2, aux2 = fun2(w, a, b)
    v3, aux3 = fun3(w, a, b)
    v4, aux4 = fun4(w, a, b)
    v5, aux5 = fun5(w, a, b)
    v6, aux6 = fun6(w, a, b)
    self._assert_items_equal((v1, v2, v3, v4, v5, v6))
    self._assert_items_equal((aux1, aux3, aux5))
    self._assert_items_equal((aux2, aux4, aux6, None))

    g1, aux1 = grad1(w, a, b)
    g2, aux2 = grad2(w, a, b)
    g3, aux3 = grad3(w, a, b)
    g4, aux4 = grad4(w, a, b)
    g5, aux5 = grad5(w, a, b)
    g6, aux6 = grad6(w, a, b)
    self._assert_items_equal((g1, g2, g3, g4, g5, g6))
    self._assert_items_equal((aux1, aux3, aux5))
    self._assert_items_equal((aux2, aux4, aux6, None))
    self.assertAllClose(aux1, aux3)

    (v1, aux1), g1 = vg1(w, a, b)
    (v2, aux2), g2 = vg2(w, a, b)
    (v3, aux3), g3 = vg3(w, a, b)
    (v4, aux4), g4 = vg4(w, a, b)
    (v5, aux5), g5 = vg5(w, a, b)
    (v6, aux6), g6 = vg6(w, a, b)
    self._assert_items_equal((v1, v2, v3, v4, v5, v6))
    self._assert_items_equal((g1, g2, g3, g4, g5, g6))
    self._assert_items_equal((aux1, aux3, aux5))
    self._assert_items_equal((aux2, aux4, aux6, None))


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Hager-Zhang line search algorithm."""

# This is based on:
# [1] W. Hager, H. Zhang, A new conjugate gradient method with guaranteed
# descent and an efficient line search. SIAM J. Optim., Vol 16. 1, pp. 170-172.
# 2005. https://www.math.lsu.edu/~hozhang/papers/cg_descent.pdf
#
# Algorithm details are from
# [2] W. Hager, H. Zhang, Algorithm 851: CG_DESCENT, a Conjugate Gradient Method
# with Guaranteed Descent.
# https://www.math.lsu.edu/~hozhang/papers/cg_compare.pdf

from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import NamedTuple
from typing import Optional
from typing import Union

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_add_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_vdot_real
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_conj


def _failed_nan(value, grad):
  return jnp.isnan(value) | jnp.isnan(grad)


class HagerZhangLineSearchState(NamedTuple):
  """Named tuple containing state information."""
  iter_num: int
  done: bool
  low: float
  high: float
  value: float
  error: float
  params: Any
  grad: Any
  failed: bool
  aux: Optional[Any] = None

  num_fun_eval: int = 0
  num_grad_eval: int = 0


@dataclass(eq=False)
class HagerZhangLineSearch(base.IterativeLineSearch):
  """Hager-Zhang line search.

  Supports complex variables.

  Attributes:
    fun: a function of the form ``fun(params, *args, **kwargs)``, where
      ``params`` are parameters of the model,
      ``*args`` and ``**kwargs`` are additional arguments.
    value_and_grad: if ``False``, ``fun`` should return the function value only.
      If ``True``, ``fun`` should return both the function value and the
      gradient.
    has_aux: if ``False``, ``fun`` should return the function value only.
      If ``True``, ``fun`` should return a pair ``(value, aux)`` where ``aux``
      is a pytree of auxiliary values.

    c1: constant used by the Wolfe and Approximate Wolfe condition.
    c2: constant strictly less than 1 used by the Wolfe and Approximate Wolfe
      condition.
    max_stepsize: upper bound on stepsize (unused).

    maxiter: maximum number of line search iterations.
    tol: tolerance of the stopping criterion.

    verbose: whether to print information on every iteration or not.

    jit: whether to JIT-compile the optimization loop (default: "auto").
    unroll: whether to unroll the optimization loop (default: "auto").
  """
  fun: Callable  # pylint:disable=g-bare-generic
  value_and_grad: Union[bool, Callable] = False
  has_aux: bool = False

  maxiter: int = 30
  tol: float = 0.

  c1: float = 0.1
  c2: float = 0.9
  expansion_factor: float = 5.0
  shrinkage_factor: float = 0.66
  approximate_wolfe_threshold = 1e-6
  # TODO(vroulet): remove max_stepsize argument as it is not used
  max_stepsize: float = 1.0 

  verbose: Union[bool, int] = False
  jit: base.AutoOrBoolean = "auto"
  unroll: base.AutoOrBoolean = "auto"

  def _value_and_grad_on_line(self, x, c, descent_direction, *args, **kwargs):
    z = tree_add_scalar_mul(x, c, descent_direction)
    (value, _), grad = self._value_and_grad_fun_with_aux(z, *args, **kwargs)
    return value, tree_vdot_real(tree_conj(grad), descent_direction)

  def _satisfies_wolfe_and_approx_wolfe(
      self,
      c,
      value_c,
      gd_vdot_c,
      value_initial,
      grad_initial,
      approx_wolfe_threshold_value,
      descent_direction):
    gd_vdot = tree_vdot_real(tree_conj(grad_initial), descent_direction)

    # Armijo condition
    # armijo = value_c <= value_initial + self.c1 * c * gd_vdot
    armijo = value_c - (value_initial + self.c1 * c * gd_vdot)
    armijo_error = jax.lax.max(armijo, 0.)
    # Curvature condition
    # curvature = gd_vdot_c >= self.c2 * gd_vdot
    curvature = self.c2 * gd_vdot - gd_vdot_c
    wolfe_error = jax.lax.max(armijo_error, curvature)

    # Approximate Wolfe
    # approx_wolfe = (2 * self.c1 - 1.) * gd_vdot >= gd_vdot_c
    approx_wolfe = gd_vdot_c - (2 * self.c1 - 1.) * gd_vdot
    approx_wolfe_error = jax.lax.max(approx_wolfe, 0.)
    approx_wolfe_error = jax.lax.max(approx_wolfe_error, curvature)
    # Finally only enable the approximate wolfe conditions when we are close in
    # value.
    approx_wolfe_error = jax.lax.max(
        approx_wolfe_error,
        (value_c - approx_wolfe_threshold_value))
    # We succeed if we either satisfy the Wolfe conditions or the approximate
    # Wolfe conditions.
    return jax.lax.min(wolfe_error, approx_wolfe_error)

  def _update(
      self, x, low, high, middle, approx_wolfe_threshold_value,
      descent_direction, fun_args, fun_kwargs):

    value_middle, grad_middle = self._value_and_grad_on_line(
        x, middle, descent_direction, *fun_args, **fun_kwargs)

    # Corresponds to the `update` subroutine in the paper.
    # This tries to create a smaller interval contained in `[low, high]`
    # from the point `middle` that satisfies the opposite slope condition, where
    # the left end point is equal to within tolerance of the initial value.

    def cond_fn(state):
      done, failed, low, middle, high, *_ = state
      return jnp.any((middle < high) & (middle > low) & ~done & ~failed)

    def body_fn(state):
      done, failed, low, middle, high, value_middle, grad_middle, it = state
      # Correspond to U1 in the paper.
      update_right_endpoint = grad_middle >= 0.
      new_high = jnp.where(~done & update_right_endpoint, middle, high)
      done = done | update_right_endpoint

      # Correspond to U2 in the paper.
      update_left_endpoint = value_middle <= approx_wolfe_threshold_value
      # Note that ~done implies grad_middle < 0. which is necessary for this
      # check.
      new_low = jnp.where(~done & update_left_endpoint, middle, low)
      done = done | update_left_endpoint

      # Correspond to U3 in the paper.
      new_high = jnp.where(~done, middle, new_high)
      done = done | jnp.isneginf(value_middle)

      # TODO(srvasude): Allow this parameter to be varied.
      new_middle = jnp.where(~done, (low + high) / 2., middle)

      new_value_middle, new_grad_middle = self._value_and_grad_on_line(
          x, new_middle, descent_direction, *fun_args, **fun_kwargs)
      new_value_middle = jnp.where(~done, new_value_middle, value_middle)
      new_grad_middle = jnp.where(~done, new_grad_middle, grad_middle)
      failed = failed | _failed_nan(new_value_middle, new_grad_middle)

      return (done,
              failed,
              new_low,
              new_middle,
              new_high,
              new_value_middle,
              new_grad_middle,
              it + 1)

    _, failed, final_low, _, final_high, _, _, nit = jax.lax.while_loop(
        cond_fn,
        body_fn,
        ((middle >= high) | (middle <= low),
         _failed_nan(value_middle, grad_middle),
         low,
         middle,
         high,
         value_middle,
         grad_middle,
         0))
    num_fun_grad_calls = nit + 1
    return failed, final_low, final_high, num_fun_grad_calls

  def _secant(self, x, low, high, descent_direction, *args, **kwargs):
    _, dlow = self._value_and_grad_on_line(
        x, low, descent_direction, *args, **kwargs)
    _, dhigh = self._value_and_grad_on_line(
        x, high, descent_direction, *args, **kwargs)
    return (low * dhigh - high * dlow) / (dhigh  - dlow)

  def _secant2(
      self, x, low, high,
      approx_wolfe_threshold_value, descent_direction, *args, **kwargs):

    # Corresponds to the secant^2 routine in the paper.

    c = self._secant(x, low, high, descent_direction, *args, **kwargs)
    num_fun_grad_calls = 2
    failed, new_low, new_high, num_fun_grad_calls_update = self._update(
        x, low, high, c, approx_wolfe_threshold_value,
        descent_direction, args, kwargs)
    num_fun_grad_calls += num_fun_grad_calls_update
    on_left_boundary = jnp.equal(c, new_low)
    on_right_boundary = jnp.equal(c, new_high)
    c = jnp.where(on_right_boundary, self._secant(
        x, high, new_high, descent_direction, *args, **kwargs), c)
    c = jnp.where(on_left_boundary, self._secant(
        x, low, new_low, descent_direction, *args, **kwargs), c)
    num_fun_grad_calls += 4

    def _reupdate():
      return self._update(
          x, new_low, new_high, c, approx_wolfe_threshold_value,
          descent_direction, args, kwargs)

    failed, new_low, new_high, num_fun_grad_calls_update = jax.lax.cond(
        on_left_boundary | on_right_boundary,
        _reupdate, lambda: (failed, new_low, new_high, 0))
    num_fun_grad_calls += num_fun_grad_calls_update
    return failed, new_low, new_high, num_fun_grad_calls

  def _bracket(
      self, x, c, approx_wolfe_threshold_value,
      descent_direction, *args, **kwargs):
    # Initial interval that satisfies the opposite slope condition.

    def cond_fn(state):
      return jnp.any(~state[0]) & ~jnp.all(state[1])

    def body_fn(state):
      (done,
       failed,
       low,
       middle,
       high,
       value_middle,
       grad_middle,
       best_middle,
       num_fun_grad_calls) = state
      # Correspond to B1 in the paper.
      update_right_endpoint = grad_middle >= 0.
      new_high = jnp.where(~done & update_right_endpoint, middle, high)
      new_low = jnp.where(~done & update_right_endpoint, best_middle, low)
      done = done | update_right_endpoint

      # Correspond to B2 in the paper.
      # Note that ~done implies grad_middle < 0. at this point so we omit
      # checking that.
      reupdate = ~done & (value_middle > approx_wolfe_threshold_value)

      def _update_interval():
        return self._update(
            x,
            0,
            middle,
            middle / 2.,
            approx_wolfe_threshold_value,
            descent_direction, args, kwargs)

      new_failed, new_low, new_high, new_num_fun_grad_calls = jax.lax.cond(
          reupdate, _update_interval, lambda: (failed, new_low, new_high, 0))
      failed = failed | new_failed
      done = done | reupdate

      # This corresponds to the largest middle value that we have probed
      # so far, that also is 'valid' (decreases the function sufficiently).
      best_middle = jnp.where(
          ~done & (value_middle <= approx_wolfe_threshold_value),
          middle, best_middle)

      # Corresponds to B3 in the paper. Increase the point and recompute.
      new_middle = jnp.where(~done, self.expansion_factor * middle, middle)

      new_value_middle, new_grad_middle = self._value_and_grad_on_line(
          x, new_middle, descent_direction, *args, **kwargs)
      new_num_fun_grad_calls += 1
      num_fun_grad_calls += new_num_fun_grad_calls

      # Terminate on encountering NaNs to avoid an infinite loop.
      failed = failed  | _failed_nan(new_value_middle, new_grad_middle)
      return (done,
              failed,
              new_low,
              new_middle,
              new_high,
              new_value_middle,
              new_grad_middle,
              best_middle,
              num_fun_grad_calls)

    value_c, grad_c = self._value_and_grad_on_line(
        x, c, descent_direction, *args, **kwargs)
    num_fun_grad_calls = 1

    # We have failed if there is a NaN at the right endpoint, or the gradient is
    # NaN at the right endpoint (when there is a finite value).
    failed = _failed_nan(value_c, grad_c)
    # If the right endpoint is -inf, then we are done as this is a minima.
    done = jnp.isneginf(value_c)

    _, failed, final_low, _, final_high, _, _, _, new_num_fun_grad_calls = jax.lax.while_loop(
        cond_fn,
        body_fn,
        (done,
         failed,
         jnp.array(0.),
         c,
         c,
         value_c,
         grad_c,
         jnp.array(0.),
         0))
    num_fun_grad_calls += new_num_fun_grad_calls
    return failed, final_low, final_high, num_fun_grad_calls

  def init_state(
      self,
      init_stepsize: float,
      params: Any,
      value: Optional[float] = None,
      grad: Optional[Any] = None,
      descent_direction: Optional[Any] = None,
      fun_args: list = [],
      fun_kwargs: dict = {},
  ) -> HagerZhangLineSearchState:
    """Initialize the line search state.

    Args:
      init_stepsize: initial step size value. This is ignored by the linesearch.
      params: current parameters.
      value: current function value (recomputed if None).
      grad: current gradient (recomputed if None).
      descent_direction: ignored.
      fun_args: additional positional arguments to be passed to ``fun``.
      fun_kwargs: additional keyword arguments to be passed to ``fun``.
    Returns:
      state
    """
    del init_stepsize

    if value is None or grad is None:
      (value, _), grad = self._value_and_grad_fun_with_aux(
          params, *fun_args, **fun_kwargs
      )
      num_fun_eval = 1
      num_grad_eval = 1
    else:
      num_fun_eval = 0
      num_grad_eval = 0


    if descent_direction is None:
      descent_direction = tree_scalar_mul(-1, tree_conj(grad))

    approx_wolfe_threshold_value = (
        value + self.approximate_wolfe_threshold * jnp.abs(value))

    # Create initial interval.
    failed, low, high, num_fun_grad_calls = self._bracket(
        params,
        jnp.ones_like(value),
        approx_wolfe_threshold_value,
        descent_direction,
        *fun_args,
        **fun_kwargs
    )
    num_fun_eval += num_fun_grad_calls
    num_grad_eval += num_fun_grad_calls

    value_low, grad_low = self._value_and_grad_on_line(
        params, low, descent_direction, *fun_args, **fun_kwargs)

    value_high, grad_high = self._value_and_grad_on_line(
        params, high, descent_direction, *fun_args, **fun_kwargs)
    num_fun_eval += 2
    num_grad_eval +=  2

    best_point = jnp.where(value_low < value_high, low, high)
    gd_vdot_best_point = jnp.where(value_low < value_high, grad_low, grad_high)
    value_best_point = jnp.minimum(value_low, value_high)

    error = self._satisfies_wolfe_and_approx_wolfe(
        best_point,
        value_best_point,
        gd_vdot_best_point,
        value,
        grad,
        approx_wolfe_threshold_value,
        descent_direction)
    done = error <= self.tol

    return HagerZhangLineSearchState(
        iter_num=jnp.asarray(0),
        low=low,
        high=high,
        error=error,
        done=done,
        value=value,
        aux=None,  # we do not need to have aux in the initial state
        params=params,
        grad=grad,
        failed=failed,
        num_fun_eval=jnp.array(num_fun_eval, base.NUM_EVAL_DTYPE),
        num_grad_eval=jnp.array(num_grad_eval, base.NUM_EVAL_DTYPE))

  def update(
      self,
      stepsize: float,
      state: NamedTuple,
      params: Any,
      value: Optional[float] = None,
      grad: Optional[Any] = None,
      descent_direction: Optional[Any] = None,
      fun_args: list = [],
      fun_kwargs: dict = {},
  ) -> base.LineSearchStep:
    """Performs one iteration of Hager-Zhang line search.

    Args:
      stepsize: current estimate of the step size.
      state: named tuple containing the line search state.
      params: current parameters.
      value: current function value (recomputed if None).
      grad: current gradient (recomputed if None).
      descent_direction: descent direction (negative gradient if None).
      fun_args: additional positional arguments to be passed to ``fun``.
      fun_kwargs: additional keyword arguments to be passed to ``fun``.
    Returns:
      (params, state)
    """
    new_num_fun_eval = state.num_fun_eval
    new_num_grad_eval = state.num_grad_eval
    if value is None or grad is None:
      (value, _), grad = self._value_and_grad_fun_with_aux(
          params, *fun_args, **fun_kwargs
      )
      new_num_fun_eval = new_num_fun_eval + 1
      new_num_grad_eval = new_num_grad_eval + 1

    if descent_direction is None:
      descent_direction = tree_scalar_mul(-1, tree_conj(grad))

    approx_wolfe_threshold_value = (
        value + self.approximate_wolfe_threshold * jnp.abs(value))

    failed, new_low, new_high, num_fun_grad_calls = self._secant2(
        params,
        state.low,
        state.high,
        approx_wolfe_threshold_value,
        descent_direction,
        *fun_args, **fun_kwargs)
    new_num_fun_eval += num_fun_grad_calls
    new_num_grad_eval += num_fun_grad_calls

    failed = state.failed | failed

    new_low = jnp.where(state.done, state.low, new_low)
    new_high = jnp.where(state.done, state.high, new_high)

    def _reupdate():
      c = (new_low + new_high) / 2.
      return self._update(
          params, new_low, new_high, c, approx_wolfe_threshold_value,
          descent_direction, fun_args, fun_kwargs)

    failed, new_low, new_high, num_fun_grad_calls = jax.lax.cond(
        ~state.done & ((new_high - new_low) >
                       (self.shrinkage_factor * (state.high - state.low))),
        _reupdate, lambda: (failed, new_low, new_high, 0))
    new_num_fun_eval += num_fun_grad_calls
    new_num_grad_eval += num_fun_grad_calls

    # Check wolfe and approximate wolfe conditions and update them.

    value_low, grad_low = self._value_and_grad_on_line(
        params, new_low, descent_direction, *fun_args, **fun_kwargs)
    value_high, grad_high = self._value_and_grad_on_line(
        params, new_high, descent_direction, *fun_args, **fun_kwargs)
    new_num_fun_eval += 2
    new_num_grad_eval += 2

    best_point = jnp.where(value_low < value_high, new_low, new_high)
    gd_vdot_best_point = jnp.where(value_low < value_high, grad_low, grad_high)

    value_best_point = jnp.minimum(value_low, value_high)

    new_stepsize = jnp.where(state.done, stepsize, best_point)
    new_params = tree_add_scalar_mul(params, best_point, descent_direction)
    (new_value, new_aux), new_grad = self._value_and_grad_fun_with_aux(
        new_params, *fun_args, **fun_kwargs
    )
    new_num_fun_eval += 1
    new_num_grad_eval += 1

    error = jnp.where(state.done, state.error,
                      self._satisfies_wolfe_and_approx_wolfe(
                          best_point,
                          value_best_point,
                          gd_vdot_best_point,
                          value,
                          grad,
                          approx_wolfe_threshold_value,
                          descent_direction))
    done = state.done | (error <= self.tol)
    failed = failed | ((state.iter_num + 1 == self.maxiter) & ~done)

    new_state = HagerZhangLineSearchState(
        iter_num=state.iter_num + 1,
        value=new_value,
        grad=new_grad,
        aux=new_aux,
        params=new_params,
        low=new_low,
        high=new_high,
        error=error,
        done=done,
        failed=failed,
        num_fun_eval=new_num_fun_eval,
        num_grad_eval=new_num_grad_eval)

    if self.verbose:
      self.log_info(
        new_state,
        error_name="Minimum Decrease & Curvature Errors",
        additional_info={
            "Stepsize": new_stepsize,
            "Objective Value": new_value
        }
      )

    return base.LineSearchStep(stepsize=new_stepsize, state=new_state)

  def __post_init__(self):
    _, _, self._value_and_grad_fun_with_aux = base._make_funs_with_aux(self.fun, self.value_and_grad, self.has_aux)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt import PolyakSGD
from nemos.third_party.jaxopt.jaxopt._src import test_util

import numpy as onp
from sklearn import datasets


class PolyakSgdTest(test_util.JaxoptTestCase):

  @parameterized.product(momentum=[0.0, 0.9], sps_variant=['SPS_max', 'SPS+'])
  def test_logreg_overparameterized(self, momentum, sps_variant):
    # Test SPS on an over-parameterized logistic regression problem.
    # The loss' infimum is zero and SPS should converge to a minimizer.
    data = datasets.make_classification(
        n_samples=10, n_features=10, random_state=0
    )
    # fun(params, data)
    fun = objective.l2_multiclass_logreg_with_intercept
    n_classes = len(jnp.unique(data[1]))

    w_init = jnp.zeros((data[0].shape[1], n_classes))
    b_init = jnp.zeros(n_classes)
    params = (w_init, b_init)

    opt = PolyakSGD(fun=fun, fun_min=0, momentum=momentum, variant=sps_variant)
    error_init = opt.l2_optimality_error(params, l2reg=0, data=data)
    params, _ = opt.run(params, l2reg=0., data=data)

    # Check optimality conditions.
    error = opt.l2_optimality_error(params, l2reg=0., data=data)
    self.assertLessEqual(error / error_init, 0.01)

  @parameterized.product(has_aux=[True, False])
  def test_logreg_with_intercept_run(self, has_aux):
    X, y = datasets.make_classification(n_samples=10, n_features=5, n_classes=3,
                                        n_informative=3, random_state=0)
    data = (X, y)
    l2reg = 100.0
    _fun = objective.l2_multiclass_logreg_with_intercept
    if has_aux:
      def fun(params, l2reg, data):
        return _fun(params, l2reg, data), None
    else:
      fun = _fun

    n_classes = len(jnp.unique(y))

    W_init = jnp.zeros((X.shape[1], n_classes))
    b_init = jnp.zeros(n_classes)
    pytree_init = (W_init, b_init)

    opt = PolyakSGD(fun=fun, max_stepsize=0.01, maxiter=300, has_aux=has_aux)
    # Test positional, keyword and mixed arguments.
    for params, _ in (opt.run(pytree_init, l2reg, data),
                      opt.run(pytree_init, l2reg=l2reg, data=data),
                      opt.run(pytree_init, l2reg, data=data)):

      # Check optimality conditions.
      error = opt.l2_optimality_error(params, l2reg=l2reg, data=data)
      self.assertLessEqual(error, 0.05)

  def test_logreg_with_intercept_run_iterable(self):
    X, y = datasets.make_classification(n_samples=10, n_features=5, n_classes=3,
                                        n_informative=3, random_state=0)

    def dataset_loader(X, y, n_iter):
      rng = onp.random.RandomState(0)
      for _ in range(n_iter):
        perm = rng.permutation(len(X))
        yield X[perm], y[perm]

    l2reg = 100.0
    fun = objective.l2_multiclass_logreg_with_intercept
    n_classes = len(jnp.unique(y))

    W_init = jnp.zeros((X.shape[1], n_classes))
    b_init = jnp.zeros(n_classes)
    pytree_init = (W_init, b_init)

    opt = PolyakSGD(fun=fun, max_stepsize=0.01, maxiter=1000)
    iterable = dataset_loader(X, y, n_iter=200)
    params, _ = opt.run_iterator(pytree_init, iterable, l2reg=l2reg)

    # Check optimality conditions.
    error = opt.l2_optimality_error(params, l2reg=l2reg, data=(X, y))
    self.assertLessEqual(error, 0.05)

  def test_logreg_autodiff(self):
    X, y = datasets.load_digits(return_X_y=True)
    data = (X, y)
    l2reg = float(X.shape[0])
    fun = objective.l2_multiclass_logreg

    jac_num = test_util.logreg_skl_jac(X, y, l2reg)
    W_skl = test_util.logreg_skl(X, y, l2reg)

    # Make sure the decorator works.
    opt = PolyakSGD(fun=fun, max_stepsize=1e-3, maxiter=5)
    def wrapper(hyperparams):
      return opt.run(W_skl, l2reg=l2reg, data=data).params
    jac_custom = jax.jacrev(wrapper)(l2reg)
    self.assertArraysAllClose(jac_num, jac_custom, atol=1e-2)

  def test_logreg_implicit_diff(self):
    X, y = datasets.load_digits(return_X_y=True)
    data = (X, y)
    l2reg = float(X.shape[0])
    fun = objective.l2_multiclass_logreg

    jac_num = test_util.logreg_skl_jac(X, y, l2reg)
    W_skl = test_util.logreg_skl(X, y, l2reg)

    # Make sure the decorator works.
    opt = PolyakSGD(fun=fun, max_stepsize=1e-3, maxiter=5, implicit_diff=True)
    def wrapper(hyperparams):
      # Unfortunately positional arguments are required when implicit_diff=True.
      return opt.run(W_skl, l2reg, data).params
    jac_custom = jax.jacrev(wrapper)(l2reg)
    self.assertArraysAllClose(jac_num, jac_custom, atol=1e-2)

if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  #jax.config.update("jax_enable_x64", True)
  absltest.main()



import dataclasses
from typing import Any, Callable, ClassVar, Type, TypeAlias

import equinox as eqx
import jax
import jax.numpy as jnp
import optimistix as optx

from ..regularizer import Regularizer
from ..typing import Pytree
from ._abstract_solver import OptimizationInfo, Params
from ._solver_adapter import SolverAdapter

DEFAULT_ATOL = 1e-8
DEFAULT_RTOL = 0.0
DEFAULT_MAX_STEPS = 100_000

OptimistixSolverState: TypeAlias = eqx.Module
OptimistixStepResult: TypeAlias = tuple[Params, OptimistixSolverState]


def _f_struct_factory():
    """Create the output shape and dtype of the objective function using the currently set JAX precision."""
    current_float_dtype = jnp.float64 if jax.config.jax_enable_x64 else jnp.float32
    return jax.ShapeDtypeStruct((), current_float_dtype)


@dataclasses.dataclass
class OptimistixConfig:
    """
    Collection of arguments required by and cached for methods of Optimistix solvers.

    They rarely need to be overwritten, and the defaults here should suffice.
    The user has the ability to overwrite them with `solver_kwargs`, and on the solver's construction
    they are saved in `OptimistixAdapter.config` for later use: passing them to `optimistix.optimise`, `init`, `step`.
    """

    # max number of steps
    maxiter: int
    # options dict passed around within optimistix
    options: dict[str, Any] = dataclasses.field(default_factory=dict)
    # "The shape+dtype of the output of `fn`"
    # dtype is dynamically determined by calling _f_struct_factory to read the currently
    # set JAX floating point precision from the config
    # f_struct: PyTree[jax.ShapeDtypeStruct] = dataclasses.field(
    f_struct: Pytree = dataclasses.field(default_factory=_f_struct_factory)
    # this would be the output shape + dtype of the aux variables fn returns
    # aux_struct: PyTree[jax.ShapeDtypeStruct] = None
    aux_struct: Pytree = None
    # "Any Lineax tags describing the structure of the Jacobian matrix d(fn)/dy."
    tags: frozenset = frozenset()
    # sets if the minimisation throws an error if an iterative solver runs out of steps
    throw: bool = False
    # norm used in the Cauchy convergence criterion. Required by all Optimistix solvers.
    norm: Callable = optx.max_norm
    # way of autodifferentiation: https://docs.kidger.site/optimistix/api/adjoints/
    adjoint: optx.AbstractAdjoint = optx.ImplicitAdjoint()
    # whether the objective function returns any auxiliary results.
    # We assume False throughout NeMoS.
    has_aux: bool = False


class OptimistixAdapter(SolverAdapter[OptimistixSolverState]):
    """
    Base class for adapters wrapping Optimistix minimizers.

    Subclasses must define the `_solver_cls` class attribute.
    The `_solver` and `stats` attributes are assumed to exist after construction,
    so if a subclass is overwriting `__init__`, these must be created.

    Note that for backward compatibility the `atol` parameter used in Optimistix
    is referred to as `tol` in NeMoS.
    """

    _solver_cls: ClassVar[Type]
    _solver: optx.AbstractMinimiser

    # used for storing info after an optimization run
    # updated with the dict from an optimistix._solution.Solution.stats
    stats: dict[str, Any]

    _proximal: ClassVar[bool] = False

    def __init__(
        self,
        unregularized_loss: Callable,
        regularizer: Regularizer,
        regularizer_strength: float | None,
        tol: float = DEFAULT_ATOL,
        rtol: float = DEFAULT_RTOL,
        maxiter: int = DEFAULT_MAX_STEPS,
        **solver_init_kwargs,
    ):
        if "atol" in solver_init_kwargs:
            raise TypeError("Please use tol instead of atol.")

        if "max_steps" in solver_init_kwargs:
            raise TypeError("Please use maxiter instead of max_steps.")

        if self._proximal:
            loss_fn = unregularized_loss
            self.prox = regularizer.get_proximal_operator()
            self.regularizer_strength = regularizer_strength
        else:
            loss_fn = regularizer.penalized_loss(
                unregularized_loss, regularizer_strength
            )
        self.fun = lambda params, args: loss_fn(params, *args)
        self.fun_with_aux = lambda params, args: (loss_fn(params, *args), None)

        # take out the arguments that go into minimise, init, terminate and so on
        # and only pass the actually needed things to __init__
        user_args = {}
        for f in dataclasses.fields(OptimistixConfig):
            kw = f.name
            if kw in solver_init_kwargs:
                user_args[kw] = solver_init_kwargs.pop(kw)
        self.config = OptimistixConfig(maxiter=maxiter, **user_args)

        self._solver = self._solver_cls(
            atol=tol,
            rtol=rtol,
            norm=self.config.norm,
            **solver_init_kwargs,
        )

        self.stats = {}

    def init_state(self, init_params: Params, *args: Any) -> OptimistixSolverState:
        return self._solver.init(
            self.fun_with_aux,
            init_params,
            args,
            self.config.options,
            self.config.f_struct,
            self.config.aux_struct,
            self.config.tags,
        )

    def update(
        self,
        params: Params,
        state: OptimistixSolverState,
        *args: Any,
    ) -> OptimistixStepResult:
        new_params, state, aux = self._solver.step(
            fn=self.fun_with_aux,
            y=params,
            args=args,
            state=state,
            options=self.config.options,
            tags=self.config.tags,
        )

        return new_params, state

    def run(
        self,
        init_params: Params,
        *args: Any,
    ) -> OptimistixStepResult:
        solution = optx.minimise(
            fn=self.fun,
            solver=self._solver,
            y0=init_params,
            args=args,
            options=self.config.options,
            has_aux=self.config.has_aux,
            max_steps=self.config.maxiter,
            adjoint=self.config.adjoint,
            throw=self.config.throw,
            tags=self.config.tags,
        )

        self.stats.update(solution.stats)

        return solution.value, solution.state

    @classmethod
    def get_accepted_arguments(cls) -> set[str]:
        own_and_solver_args = super().get_accepted_arguments()

        # atol is added from wrapped optimistix solvers
        # but currently throughout nemos tol is used
        own_and_solver_args.remove("atol")

        common_optx_arguments = set(
            [f.name for f in dataclasses.fields(OptimistixConfig)]
        )
        all_arguments = own_and_solver_args | common_optx_arguments

        return all_arguments

    @classmethod
    def _note_about_accepted_arguments(cls) -> str:
        return """
        Note that for backward compatibility the `atol` parameter used in Optimistix
        is referred to as `tol` in NeMoS.
        """

    @property
    def maxiter(self) -> int:
        return self.config.maxiter

    def get_optim_info(self, state: OptimistixSolverState) -> OptimizationInfo:
        num_steps = self.stats["num_steps"].item()

        function_val = (
            state.f.item() if hasattr(state, "f") else state.f_info.f.item()
        )  # pyright: ignore

        return OptimizationInfo(
            function_val=function_val,
            num_steps=num_steps,
            converged=state.terminate.item(),  # pyright: ignore
            reached_max_steps=(num_steps == self.maxiter),
        )


class OptimistixBFGS(OptimistixAdapter):
    """Adapter for optimistix.BFGS."""

    _solver_cls = optx.BFGS


class OptimistixNonlinearCG(OptimistixAdapter):
    """Adapter for optimistix.NonlinearCG."""

    _solver_cls = optx.NonlinearCG



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt._src import linear_solve as _linear_solve
from nemos.third_party.jaxopt.jaxopt import linear_solve
from nemos.third_party.jaxopt.jaxopt._src import test_util

import numpy as onp


class LinearSolveTest(test_util.JaxoptTestCase):

  def test_materialize_array(self):
    rng = onp.random.RandomState(0)

    # Matrix case.
    A = rng.randn(5, 5)
    matvec = lambda x: jnp.dot(A, x)
    A2 = _linear_solve._materialize_array(matvec, (5,))
    self.assertArraysAllClose(A, A2)

    # Tensor case.
    A = rng.randn(5, 3, 5, 3)
    A_mat = A.reshape(15, 15)
    matvec = lambda x: jnp.dot(A_mat, x.ravel()).reshape(5, 3)
    A2 = _linear_solve._materialize_array(matvec, (5, 3))
    self.assertArraysAllClose(A, A2, atol=1e-3)

  def test_rmatvec(self):
    rng = onp.random.RandomState(0)

    # Test square case.
    A = rng.randn(5, 5)
    matvec = lambda x: jnp.dot(A, x)
    x = rng.randn(5)
    rmatvec = _linear_solve._make_rmatvec(matvec, x)
    self.assertArraysAllClose(rmatvec(x), jnp.dot(A.T, x))

    # Test non-square case
    A = rng.randn(5, 3)
    matvec = lambda x: jnp.dot(A, x)
    y = rng.randn(5)
    example_x = rng.randn(3)
    rmatvec = _linear_solve._make_rmatvec(matvec, example_x)
    self.assertArraysAllClose(rmatvec(y), jnp.dot(A.T, y))

  def test_normal_matvec(self):
    rng = onp.random.RandomState(0)
    A = rng.randn(5, 5)
    matvec = lambda x: jnp.dot(A, x)
    x = rng.randn(5)
    self.assertArraysAllClose(_linear_solve._normal_matvec(matvec, x),
                              jnp.dot(A.T, jnp.dot(A, x)))

  def test_solve_dense(self):
    rng = onp.random.RandomState(0)

    # Matrix case.
    A = rng.randn(5, 5)
    b = rng.randn(5)
    matvec = lambda x: jnp.dot(A, x)
    x = linear_solve.solve_lu(matvec, b)
    x2 = jax.numpy.linalg.solve(A, b)
    x3 = linear_solve.solve_iterative_refinement(matvec, b)
    x4 = linear_solve.solve_qr(matvec, b)
    self.assertArraysAllClose(x, x2)
    self.assertArraysAllClose(x, x3)
    self.assertArraysAllClose(x, x4, atol=1e-4)

    # Tensor case.
    A = rng.randn(5, 3, 5, 3)
    A_mat = A.reshape(15, 15)
    b = rng.randn(5, 3)

    def matvec(x):
      return jnp.dot(A_mat, x.ravel()).reshape(5, 3)

    x = linear_solve.solve_lu(matvec, b)
    x2 = linear_solve.solve_gmres(matvec, b)
    x3 = linear_solve.solve_iterative_refinement(matvec, b)
    self.assertArraysAllClose(x, x2, atol=1e-4)
    self.assertArraysAllClose(x, x3, atol=1e-4)

  def test_solve_dense_psd(self):
    rng = onp.random.RandomState(0)
    A = rng.randn(5, 5)
    A = jnp.dot(A.T, A)
    b = rng.randn(5)
    matvec = lambda x: jnp.dot(A, x)
    x = linear_solve.solve_cholesky(matvec, b)
    x2 = jax.numpy.linalg.solve(A, b)
    x3 = linear_solve.solve_inv(matvec, b)
    x4 = linear_solve.solve_qr(matvec, b)
    self.assertArraysAllClose(x, x2, atol=1e-2)
    self.assertArraysAllClose(x, x3, atol=1e-2)
    self.assertArraysAllClose(x, x4, atol=1e-2)

  def test_solve_sparse(self):
    rng = onp.random.RandomState(0)

    # Matrix case.
    A = rng.randn(5, 5)
    b = rng.randn(5)

    def matvec(x):
      return jnp.dot(A, x)

    x = linear_solve.solve_lu(matvec, b)
    x2 = linear_solve.solve_normal_cg(matvec, b)
    x3 = linear_solve.solve_gmres(matvec, b)
    x4 = linear_solve.solve_bicgstab(matvec, b)
    x5 = linear_solve.solve_iterative_refinement(matvec, b)
    x6 = linear_solve.solve_qr(matvec, b)
    
    self.assertArraysAllClose(x, x2, atol=1e-4, rtol=1e-4)
    self.assertArraysAllClose(x, x3, atol=1e-4, rtol=1e-4)
    self.assertArraysAllClose(x, x4, atol=1e-4, rtol=1e-4)
    self.assertArraysAllClose(x, x5, atol=1e-4, rtol=1e-4)
    self.assertArraysAllClose(x, x6, atol=1e-4, rtol=1e-4)

  def test_solve_sparse_ridge(self):
    rng = onp.random.RandomState(0)

    # Value of the ridge regularizaer.
    ridge = 1.0

    # Matrix case.
    A = rng.randn(5, 5)
    b = rng.randn(5)

    def matvec(x):
      return jnp.dot(A, x)

    def matvec_with_ridge(x):
      return jnp.dot(A + ridge * onp.eye(A.shape[0]), x)

    x = linear_solve.solve_lu(matvec_with_ridge, b)
    x2 = linear_solve.solve_gmres(matvec, b, ridge=ridge)
    x3 = linear_solve.solve_bicgstab(matvec, b, ridge=ridge)
    x4 = linear_solve.solve_inv(matvec, b, ridge=ridge)
    self.assertArraysAllClose(x, x2, atol=1e-4)
    self.assertArraysAllClose(x, x3, atol=1e-4)
    self.assertArraysAllClose(x, x4, atol=1e-4)

    K = onp.dot(A.T, A)
    Ab = onp.dot(A.T, b)

    def gram_matvec(x):
      return jnp.dot(K, x)

    def gram_matvec_with_ridge(x):
      return jnp.dot(K + ridge * onp.eye(A.shape[0]), x)

    x = linear_solve.solve_lu(gram_matvec_with_ridge, b)
    x4 = linear_solve.solve_cg(gram_matvec, b, ridge=ridge)
    self.assertArraysAllClose(x, x4, atol=1e-4)

    x = linear_solve.solve_lu(gram_matvec_with_ridge, Ab)
    x5 = linear_solve.solve_normal_cg(matvec, b, ridge=ridge)
    self.assertArraysAllClose(x, x5, atol=1e-4)

  def test_solve_1d(self):
    rng = onp.random.RandomState(0)

    A = rng.randn(1, 1)
    b = rng.randn(1)

    def matvec(x):
      return jnp.dot(A, x)

    x = linear_solve.solve_lu(matvec, b)
    self.assertArraysAllClose(x, b / A[0, 0])

  def test_solve_normal_cg_nonsquare(self):
    rng = onp.random.RandomState(0)

    A = rng.randn(2, 3)
    b = rng.randn(2)

    def matvec(x):
      return jnp.dot(A, x)

    def normal_matvec(x):
      return jnp.dot(A.T, jnp.dot(A, x))

    self.assertRaises(TypeError, linear_solve.solve_normal_cg, matvec, b)

    x = linear_solve.solve_normal_cg(matvec, b, init=jnp.zeros(3))
    x2 = linear_solve.solve_cg(normal_matvec, jnp.dot(A.T, b))
    self.assertArraysAllClose(x, x2)


if __name__ == '__main__':
  absltest.main()
